#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_isis_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

"""
test_isis_topo1.py: Test ISIS topology.
"""
import time
import datetime
import functools
import json
import os
import re
import sys
import pytest

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.common_config import (
    retry,
    stop_router,
    start_router,
)
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger


pytestmark = [pytest.mark.isisd]

VERTEX_TYPE_LIST = [
    "pseudo_IS",
    "pseudo_TE-IS",
    "IS",
    "TE-IS",
    "ES",
    "IP internal",
    "IP external",
    "IP TE",
    "IP6 internal",
    "IP6 external",
    "UNKNOWN",
]


def build_topo(tgen):
    "Build function"

    # Add ISIS routers:
    # r1      r2
    #  | sw1  | sw2
    # r3     r4
    #  |      |
    # sw3    sw4
    #   \    /
    #     r5
    for routern in range(1, 6):
        tgen.add_router("r{}".format(routern))

    # r1 <- sw1 -> r3
    sw = tgen.add_switch("sw1")
    sw.add_link(tgen.gears["r1"])
    sw.add_link(tgen.gears["r3"])

    # r2 <- sw2 -> r4
    sw = tgen.add_switch("sw2")
    sw.add_link(tgen.gears["r2"])
    sw.add_link(tgen.gears["r4"])

    # r3 <- sw3 -> r5
    sw = tgen.add_switch("sw3")
    sw.add_link(tgen.gears["r3"])
    sw.add_link(tgen.gears["r5"])

    # r4 <- sw4 -> r5
    sw = tgen.add_switch("sw4")
    sw.add_link(tgen.gears["r4"])
    sw.add_link(tgen.gears["r5"])


def setup_module(mod):
    "Sets up the pytest environment"
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    # For all registered routers, load the zebra configuration file
    for rname, router in tgen.routers().items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
        )

    # After loading the configurations, this function loads configured daemons.
    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"
    tgen = get_topogen()

    # This function tears down the whole topology.
    tgen.stop_topology()


def test_isis_convergence():
    "Wait for the protocol to converge before starting to test"
    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("waiting for ISIS protocol to converge")
    for rname, router in tgen.routers().items():
        filename = "{0}/{1}/{1}_topology.json".format(CWD, rname)
        expected = json.loads(open(filename).read())

        def compare_isis_topology(router, expected):
            "Helper function to test ISIS topology convergence."
            actual = json.loads(router.vtysh_cmd("show isis topology json"))
            return topotest.json_cmp(actual, expected)

        test_func = functools.partial(compare_isis_topology, router, expected)
        (result, diff) = topotest.run_and_expect(test_func, None, wait=0.5, count=120)
        assert result, "ISIS did not converge on {}:\n{}".format(rname, diff)


def test_isis_route_installation():
    "Check whether all expected routes are present"
    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking routers for installed ISIS routes")

    # Check for routes in 'show ip route json'
    for rname, router in tgen.routers().items():
        filename = "{0}/{1}/{1}_route.json".format(CWD, rname)
        expected = json.loads(open(filename, "r").read())

        def compare_isis_installed_routes(router, expected):
            "Helper function to test ISIS routes installed in rib."
            actual = router.vtysh_cmd("show ip route json", isjson=True)
            return topotest.json_cmp(actual, expected)

        test_func = functools.partial(compare_isis_installed_routes, router, expected)
        (result, _) = topotest.run_and_expect(test_func, None, wait=1, count=10)
        assertmsg = "Router '{}' routes mismatch".format(rname)
        assert result, assertmsg


def test_isis_linux_route_installation():
    "Check whether all expected routes are present and installed in the OS"
    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking routers for installed ISIS routes in OS")

    # Check for routes in `ip route`
    for rname, router in tgen.routers().items():
        filename = "{0}/{1}/{1}_route_linux.json".format(CWD, rname)
        expected = json.loads(open(filename, "r").read())
        actual = topotest.ip4_route(router)
        assertmsg = "Router '{}' OS routes mismatch".format(rname)
        assert topotest.json_cmp(actual, expected) is None, assertmsg


def test_isis_route6_installation():
    "Check whether all expected routes are present"
    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking routers for installed ISIS IPv6 routes")

    # Check for routes in 'show ip route json'
    for rname, router in tgen.routers().items():
        filename = "{0}/{1}/{1}_route6.json".format(CWD, rname)
        expected = json.loads(open(filename, "r").read())

        def compare_isis_v6_installed_routes(router, expected):
            "Helper function to test ISIS v6 routes installed in rib."
            actual = router.vtysh_cmd("show ipv6 route json", isjson=True)
            return topotest.json_cmp(actual, expected)

        test_func = functools.partial(
            compare_isis_v6_installed_routes, router, expected
        )
        (result, _) = topotest.run_and_expect(test_func, None, wait=1, count=10)
        assertmsg = "Router '{}' routes mismatch".format(rname)
        assert result, assertmsg


def test_isis_linux_route6_installation():
    "Check whether all expected routes are present and installed in the OS"
    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking routers for installed ISIS IPv6 routes in OS")

    # Check for routes in `ip route`
    for rname, router in tgen.routers().items():
        filename = "{0}/{1}/{1}_route6_linux.json".format(CWD, rname)
        expected = json.loads(open(filename, "r").read())
        actual = topotest.ip6_route(router)
        assertmsg = "Router '{}' OS routes mismatch".format(rname)
        assert topotest.json_cmp(actual, expected) is None, assertmsg


def test_isis_summary_json():
    "Check json struct in show isis summary json"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking 'show isis summary json'")
    for rname, _ in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd("show isis summary json", isjson=True)
        assertmsg = "Test isis summary json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert json_output["vrfs"][0]["vrf"] == "default", assertmsg
        assert json_output["vrfs"][0]["areas"][0]["area"] == "1", assertmsg
        assert json_output["vrfs"][0]["areas"][0]["levels"][0]["id"] != "3", assertmsg


def test_isis_interface_json():
    "Check json struct in show isis interface json"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking 'show isis interface json'")
    for rname, _ in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd(
            "show isis interface json", isjson=True
        )
        assertmsg = "Test isis interface json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert (
            json_output["areas"][0]["circuits"][0]["interface"]["name"]
            == rname + "-eth0"
        ), assertmsg

    for rname, router in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd(
            "show isis interface detail json", isjson=True
        )
        assertmsg = "Test isis interface json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert (
            json_output["areas"][0]["circuits"][0]["interface"]["name"]
            == rname + "-eth0"
        ), assertmsg


def test_isis_neighbor_json():
    "Check json struct in show isis neighbor json"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # tgen.mininet_cli()
    logger.info("Checking 'show isis neighbor json'")
    for rname, _ in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd(
            "show isis neighbor json", isjson=True
        )
        assertmsg = "Test isis neighbor json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert (
            json_output["areas"][0]["circuits"][0]["interface"] == rname + "-eth0"
        ), assertmsg

    for rname, router in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd(
            "show isis neighbor detail json", isjson=True
        )
        assertmsg = "Test isis neighbor json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert (
            json_output["areas"][0]["circuits"][0]["interface"]["name"]
            == rname + "-eth0"
        ), assertmsg


def test_isis_neighbor_state():
    "Check that the neighbor states remain normal when the ISIS type is switched."

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Checking 'show isis neighbor state on a p2p link'")

    # Establish a P2P link
    # When the IS-IS type of r3 is set to level-1-2 and the IS-IS type of r5 is set to level-1,
    # it is expected that all neighbors exist and are in the Up state
    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        """
        configure
        router isis 1
            no redistribute ipv4 connected level-1
            no redistribute ipv4 connected level-2
            no redistribute ipv6 connected level-1
            no redistribute ipv6 connected level-2
        interface r3-eth1
            no isis circuit-type
            isis network point-to-point
        end
        """
    )
    r5 = tgen.gears["r5"]
    r5.vtysh_cmd(
        """
        configure
        router isis 1
            no redistribute ipv4 connected level-1
            no redistribute ipv6 connected level-1
            no redistribute ipv4 table 20 level-1
        interface r5-eth0
            no isis circuit-type
            isis network point-to-point
        end
        """
    )
    result = _check_isis_neighbor_json("r3", "r5", True, "Up")
    assert result is True, result
    result = _check_isis_neighbor_json("r5", "r3", True, "Up")
    assert result is True, result

    # Remove the configuration that affects the switch of IS-IS type.
    # Configure the IS-IS type of r3 to transition from level-1-2 to level-2-only,
    # while maintaining the IS-IS type of r5 as level-1.
    # In this scenario,
    # the expectation is that some neighbors do not exist or are in the Initializing state
    r3.vtysh_cmd(
        """
        configure
        router isis 1
            is-type level-2-only
        end
        """
    )
    result = _check_isis_neighbor_json("r3", "r5", False, "Initializing")
    assert result is True, result
    result = _check_isis_neighbor_json("r5", "r3", False, "Initializing")
    assert result is True, result

    # Restore to initial configuration
    logger.info("Checking 'restore to initial configuration'")
    r3.vtysh_cmd(
        """
        configure
        interface r3-eth1
            isis circuit-type level-1
            no isis network point-to-point
        router isis 1
            no is-type
            redistribute ipv4 connected level-1
            redistribute ipv4 connected level-2
            redistribute ipv6 connected level-1
            redistribute ipv6 connected level-2
        end
        """
    )
    r5.vtysh_cmd(
        """
        configure
        interface r5-eth0
            isis circuit-type level-1
            no isis network point-to-point
        router isis 1
            redistribute ipv4 connected level-1
            redistribute ipv6 connected level-1
            redistribute ipv4 table 20 level-1
        end
        """
    )
    result = _check_isis_neighbor_json("r3", "r5", True, "Up")
    assert result is True, result
    result = _check_isis_neighbor_json("r5", "r3", True, "Up")
    assert result is True, result


def test_isis_database_json():
    "Check json struct in show isis database json"

    tgen = get_topogen()
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # tgen.mininet_cli()
    logger.info("Checking 'show isis database json'")
    for rname, _ in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd(
            "show isis database json", isjson=True
        )
        assertmsg = "Test isis database json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert json_output["areas"][0]["area"]["name"] == "1", assertmsg
        assert json_output["areas"][0]["levels"][0]["id"] != "3", assertmsg

    for rname, router in tgen.routers().items():
        logger.info("Checking router %s", rname)
        json_output = tgen.gears[rname].vtysh_cmd(
            "show isis database detail json", isjson=True
        )
        assertmsg = "Test isis database json failed in '{}' data '{}'".format(
            rname, json_output
        )
        assert json_output["areas"][0]["area"]["name"] == "1", assertmsg
        assert json_output["areas"][0]["levels"][0]["id"] != "3", assertmsg


def test_isis_overload_on_startup():
    "Check that overload on startup behaves as expected"

    tgen = get_topogen()
    net = get_topogen().net
    overload_time = 120

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Testing overload on startup behavior")

    # Configure set-overload-bit on-startup on r3
    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        f"""
          configure
            router isis 1
              set-overload-bit on-startup {overload_time}
        """
    )
    # Restart r3
    logger.info("Stop router")
    stop_router(tgen, "r3")
    logger.info("Start router")

    tstamp_before_start_router = datetime.datetime.now()
    start_router(tgen, "r3")
    tstamp_after_start_router = datetime.datetime.now()
    startup_router_time = (
        tstamp_after_start_router - tstamp_before_start_router
    ).total_seconds()

    # Check that the overload bit is set in r3's LSP
    check_lsp_overload_bit("r3", "r3.00-00", "0/0/1")
    check_lsp_overload_bit("r1", "r3.00-00", "0/0/1")

    # Attempt to unset overload bit while timer is still running
    r3.vtysh_cmd(
        """
          configure
            router isis 1
              no set-overload-bit on-startup
              no set-overload-bit
        """
    )

    # Check overload bit is still set
    check_lsp_overload_bit("r1", "r3.00-00", "0/0/1")

    # Check that overload bit is unset after timer completes
    check_lsp_overload_bit("r3", "r3.00-00", "0/0/0")
    tstamp_after_bit_unset = datetime.datetime.now()
    check_lsp_overload_bit("r1", "r3.00-00", "0/0/0")

    # Collect time overloaded
    time_overloaded = (
        tstamp_after_bit_unset - tstamp_after_start_router
    ).total_seconds()
    logger.info(f"Time Overloaded: {time_overloaded}")

    # Use time it took to startup router as lower bound
    logger.info(
        f"Assert that overload time falls in range: {overload_time - startup_router_time} < {time_overloaded} <= {overload_time}"
    )
    result = overload_time - startup_router_time < time_overloaded <= overload_time
    assert result


def test_isis_overload_on_startup_cancel_timer():
    "Check that overload on startup timer is cancelled when overload bit is set/unset"

    tgen = get_topogen()
    net = get_topogen().net
    overload_time = 90

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info(
        "Testing overload on startup behavior with set overload bit: cancel timer"
    )

    # Configure set-overload-bit on-startup on r3
    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        f"""
          configure
            router isis 1
              set-overload-bit on-startup {overload_time}
              set-overload-bit
        """
    )
    # Restart r3
    logger.info("Stop router")
    stop_router(tgen, "r3")
    logger.info("Start router")
    start_router(tgen, "r3")

    # Check that the overload bit is set in r3's LSP
    check_lsp_overload_bit("r3", "r3.00-00", "0/0/1")

    # Check that overload timer is running
    check_overload_timer("r3", True)

    # Unset overload bit while timer is running
    r3.vtysh_cmd(
        """
          configure
            router isis 1
              no set-overload-bit
        """
    )

    # Check that overload timer is cancelled
    check_overload_timer("r3", False)

    # Check overload bit is unset
    check_lsp_overload_bit("r3", "r3.00-00", "0/0/0")


def test_isis_overload_on_startup_override_timer():
    "Check that overload bit remains set after overload timer expires if overload bit is configured"

    tgen = get_topogen()
    net = get_topogen().net
    overload_time = 60

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info(
        "Testing overload on startup behavior with set overload bit: override timer"
    )

    # Configure set-overload-bit on-startup on r3
    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        f"""
          configure
            router isis 1
              set-overload-bit on-startup {overload_time}
              set-overload-bit
        """
    )
    # Restart r3
    logger.info("Stop router")
    stop_router(tgen, "r3")
    logger.info("Start router")
    start_router(tgen, "r3")

    # Check that the overload bit is set in r3's LSP
    check_lsp_overload_bit("r3", "r3.00-00", "0/0/1")

    # Check that overload timer is running
    check_overload_timer("r3", True)

    # Check that overload timer expired
    check_overload_timer("r3", False)

    # Check overload bit is still set
    check_lsp_overload_bit("r3", "r3.00-00", "0/0/1")


def test_isis_advertise_passive_only():
    """Check that we only advertise prefixes of passive interfaces when advertise-passive-only is enabled."""
    tgen = get_topogen()
    net = get_topogen().net
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Testing isis advertise-passive-only behavior")
    expected_prefixes_no_advertise_passive_only = set(
        ["10.0.20.0/24", "10.254.0.1/32", "2001:db8:f::1/128", "2001:db8:1:1::/64"]
    )
    expected_prefixes_advertise_passive_only = set(
        ["10.254.0.1/32", "2001:db8:f::1/128"]
    )
    lsp_id = "r1.00-00"

    r1 = tgen.gears["r1"]
    r1.vtysh_cmd(
        """
        configure
        router isis 1
         no redistribute ipv4 connected level-2
         no redistribute ipv6 connected level-2
        interface lo
         ip router isis 1
         ipv6 router isis 1
         isis passive
        end
        """
    )

    result = check_advertised_prefixes(
        r1, lsp_id, expected_prefixes_no_advertise_passive_only
    )
    assert result is True, result

    r1.vtysh_cmd(
        """
        configure
        router isis 1
         advertise-passive-only
        end
        """
    )

    result = check_advertised_prefixes(
        r1, lsp_id, expected_prefixes_advertise_passive_only
    )
    assert result is True, result


def test_isis_hello_padding_during_adjacency_formation():
    """Check that IIH packets is only padded when adjacency is still being formed
    when isis hello padding during-adjacency-formation is configured
    """
    tgen = get_topogen()
    net = get_topogen().net
    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    logger.info("Testing isis hello padding during-adjacency-formation behavior")
    r3 = tgen.gears["r3"]

    # Reduce hello-multiplier to make the adjacency go down faster.
    r3.vtysh_cmd(
        """
        configure
        interface r3-eth0
            isis hello-multiplier 2
        """
    )

    r1 = tgen.gears["r1"]
    cmd_output = r1.vtysh_cmd(
        """
        configure
        interface r1-eth0
            isis hello padding during-adjacency-formation
        end
        debug isis adj-packets
        """
    )
    result = check_last_iih_packet_for_padding(r1, expect_padding=False)
    assert result is True, result

    r3.vtysh_cmd(
        """
        configure
        interface r3-eth0
            shutdown
        """
    )
    result = check_last_iih_packet_for_padding(r1, expect_padding=True)
    assert result is True, result

    r3 = tgen.gears["r3"]
    r3.vtysh_cmd(
        """
        configure
        interface r3-eth0
            no shutdown
        """
    )
    result = check_last_iih_packet_for_padding(r1, expect_padding=False)
    assert result is True, result


def _check_isis_neighbor_json(
    self, neighbor, neighbor_expected, neighbor_state_expected
):
    tgen = get_topogen()
    router = tgen.gears[self]
    logger.info(
        f"check_isis_neighbor_json {router} {neighbor} {neighbor_expected} {neighbor_state_expected}"
    )

    result = _check_isis_neighbor_exist(self, neighbor)
    if result == True:
        return _check_isis_neighbor_state(self, neighbor, neighbor_state_expected)
    elif neighbor_expected == True:
        return "{} with expected neighbor {} got none ".format(router.name, neighbor)
    else:
        return True


@retry(retry_timeout=60)
def _check_isis_neighbor_exist(self, neighbor):
    tgen = get_topogen()
    router = tgen.gears[self]
    logger.info(f"check_isis_neighbor_exist {router} {neighbor}")
    neighbor_json = router.vtysh_cmd("show isis neighbor json", isjson=True)

    circuits = neighbor_json.get("areas", [])[0].get("circuits", [])
    for circuit in circuits:
        if "adj" in circuit and circuit["adj"] == neighbor:
            return True

    return "The neighbor {} of router {} has not been learned yet ".format(
        neighbor, router.name
    )


@retry(retry_timeout=5)
def _check_isis_neighbor_state(self, neighbor, neighbor_state_expected):
    tgen = get_topogen()
    router = tgen.gears[self]
    logger.info(
        f"check_isis_neighbor_state {router} {neighbor} {neighbor_state_expected}"
    )
    neighbor_json = router.vtysh_cmd(
        "show isis neighbor {} json".format(neighbor), isjson=True
    )

    circuits = neighbor_json.get("areas", [])[0].get("circuits", [])
    for circuit in circuits:
        interface = circuit.get("interface", {})
        if "state" in interface:
            neighbor_state = interface["state"]
            if neighbor_state == neighbor_state_expected:
                return True

    return "{} peer with expected neighbor_state {} got {} ".format(
        router.name, neighbor_state_expected, neighbor_state
    )


@retry(retry_timeout=10)
def check_last_iih_packet_for_padding(router, expect_padding):
    logfilename = "{}/{}".format(router.gearlogdir, "isisd.log")
    last_hello_packet_line = None
    with open(logfilename, "r") as f:
        lines = f.readlines()
        for line in lines:
            if re.search("Sending .+? IIH", line):
                last_hello_packet_line = line

    if last_hello_packet_line is None:
        return "Expected IIH packet in {}, but no packet found".format(logfilename)

    interface_name, packet_length = re.search(
        r"Sending .+ IIH on (.+), length (\d+)", last_hello_packet_line
    ).group(1, 2)
    packet_length = int(packet_length)
    interface_output = router.vtysh_cmd("show interface {} json".format(interface_name))
    interface_json = json.loads(interface_output)
    padded_packet_length = interface_json[interface_name]["mtu"] - 3
    if expect_padding:
        if packet_length == padded_packet_length:
            return True
        return (
            "Expected padded packet with length {}, got packet with length {}".format(
                padded_packet_length, packet_length
            )
        )
    if packet_length < padded_packet_length:
        return True
    return "Expected unpadded packet with length less than {}, got packet with length {}".format(
        padded_packet_length, packet_length
    )


@retry(retry_timeout=5)
def check_advertised_prefixes(router, lsp_id, expected_prefixes):
    output = router.vtysh_cmd("show isis database detail {}".format(lsp_id))
    prefixes = set(re.findall(r"IP(?:v6)? Reachability: (.*) \(Metric: 10\)", output))
    if prefixes == expected_prefixes:
        return True
    return str({"expected_prefixes:": expected_prefixes, "prefixes": prefixes})


@retry(retry_timeout=200)
def _check_lsp_overload_bit(router, overloaded_router_lsp, att_p_ol_expected):
    "Verfiy overload bit in router's LSP"

    tgen = get_topogen()
    router = tgen.gears[router]
    logger.info(f"check_overload_bit {router}")
    isis_database_output = router.vtysh_cmd(
        "show isis database {} json".format(overloaded_router_lsp)
    )

    database_json = json.loads(isis_database_output)
    if "lsps" not in database_json["areas"][0]["levels"][1]:
        return "The LSP of {} has not been synchronized yet ".format(router.name)

    att_p_ol = database_json["areas"][0]["levels"][1]["lsps"][0]["attPOl"]
    if att_p_ol == att_p_ol_expected:
        return True
    return "{} peer with expected att_p_ol {} got {} ".format(
        router.name, att_p_ol_expected, att_p_ol
    )


def check_lsp_overload_bit(router, overloaded_router_lsp, att_p_ol_expected):
    "Verfiy overload bit in router's LSP"

    assertmsg = _check_lsp_overload_bit(
        router, overloaded_router_lsp, att_p_ol_expected
    )
    assert assertmsg is True, assertmsg


@retry(retry_timeout=200)
def _check_overload_timer(router, timer_expected):
    "Verfiy overload bit in router's LSP"

    tgen = get_topogen()
    router = tgen.gears[router]
    output = router.vtysh_cmd("show event timers")

    timer_running = "set_overload_on_start_timer" in output
    if timer_running == timer_expected:
        return True
    return "Expected timer running status: {}".format(timer_expected)


def check_overload_timer(router, timer_expected):
    "Verfiy overload bit in router's LSP"

    assertmsg = _check_overload_timer(router, timer_expected)
    assert assertmsg is True, assertmsg


def test_memory_leak():
    "Run the memory leak test and report results."
    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))


#
# Auxiliary functions
#


def dict_merge(dct, merge_dct):
    """
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None

    Source:
    https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
    """
    for k, _ in merge_dct.items():
        if k in dct and isinstance(dct[k], dict) and topotest.is_mapping(merge_dct[k]):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def parse_topology(lines, level):
    """
    Parse the output of 'show isis topology level-X' into a Python dict.
    """
    areas = {}
    area = None
    ipv = None
    vertex_type_regex = "|".join(VERTEX_TYPE_LIST)

    for line in lines:
        area_match = re.match(r"Area (.+):", line)
        if area_match:
            area = area_match.group(1)
            if area not in areas:
                areas[area] = {level: {"ipv4": [], "ipv6": []}}
            ipv = None
            continue
        elif area is None:
            continue

        if re.match(r"IS\-IS paths to level-. routers that speak IPv6", line):
            ipv = "ipv6"
            continue
        if re.match(r"IS\-IS paths to level-. routers that speak IP", line):
            ipv = "ipv4"
            continue

        item_match = re.match(
            r"([^\s]+) ([^\s]+) ([^\s]+) ([^\s]+) ([^\s]+) ([^\s]+)", line
        )
        if (
            item_match is not None
            and item_match.group(1) == "Vertex"
            and item_match.group(2) == "Type"
            and item_match.group(3) == "Metric"
            and item_match.group(4) == "Next-Hop"
            and item_match.group(5) == "Interface"
            and item_match.group(6) == "Parent"
        ):
            # Skip header
            continue

        item_match = re.match(
            r"([^\s]+) ({}) ([0]|([1-9][0-9]*)) ([^\s]+) ([^\s]+) ([^\s]+)".format(
                vertex_type_regex
            ),
            line,
        )
        if item_match is not None:
            areas[area][level][ipv].append(
                {
                    "vertex": item_match.group(1),
                    "type": item_match.group(2),
                    "metric": item_match.group(3),
                    "next-hop": item_match.group(5),
                    "interface": item_match.group(6),
                    "parent": item_match.group(7),
                }
            )
            continue

        item_match = re.match(
            r"([^\s]+) ({}) ([0]|([1-9][0-9]*)) ([^\s]+)".format(vertex_type_regex),
            line,
        )

        if item_match is not None:
            areas[area][level][ipv].append(
                {
                    "vertex": item_match.group(1),
                    "type": item_match.group(2),
                    "metric": item_match.group(3),
                    "parent": item_match.group(5),
                }
            )
            continue

        item_match = re.match(r"([^\s]+)", line)
        if item_match is not None:
            areas[area][level][ipv].append({"vertex": item_match.group(1)})
            continue

    return areas



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# test_cspf_topo1.py
# Part of NetDEF Topology Tests
#
# Copyright (c) 2022 by Orange
# Author: Olivier Dugeon <olivier.dugeon@orange.com>
#

"""
test_cspf_topo1.py: Test the FRR Constraint Shortest Path First algorithm.

           +------------+
           |            |
           |     R1     |
           | 10.0.225.1 |
           |            |
           +------------+
        r1-eth0|    |r1-eth1
               |    |
    10.0.0.0/24|    |10.0.1.0/24
               |    |2001:db8:1:/64
               |    |
        r2-eth0|    |r2-eth1
           +------------+                  +------------+
           |            |                  |            |
           |     R2     |r2-eth2    r3-eth0|     R3     |
           | 10.0.255.2 +------------------+ 10.0.255.3 |
           |            |     10.0.3.0/24  |            |
           +------------+  2001:db8:3:/64  +------+-----+
           r2-eth3|                        r3-eth1|
                  |                               |
       10.0.4.0/24|                               |
                  |                               |
                  |                               |
           r4-eth0|                 2001:db8:5:/64|
           +------------+                         |
           |            |                         |
           |     R4     |r4-eth1                  |
           | 10.0.255.4 +-------------------------+
           |            |
           +------------+

"""

import os
import sys
import json
from functools import partial

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413

# Import topogen and topotest helpers
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.topolog import logger

# and Finally pytest
import pytest

pytestmark = [pytest.mark.isisd]


def build_topo(tgen):
    "Build function"

    # Create 4 routers
    for routern in range(1, 5):
        tgen.add_router("r{}".format(routern))

    # Interconect router 1 and 2 with 2 links
    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])
    switch = tgen.add_switch("s2")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 3 and 2
    switch = tgen.add_switch("s3")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["r2"])

    # Interconect router 4 and 2
    switch = tgen.add_switch("s4")
    switch.add_link(tgen.gears["r4"])
    switch.add_link(tgen.gears["r2"])

    # Interconnect router 3 and 4
    switch = tgen.add_switch("s5")
    switch.add_link(tgen.gears["r3"])
    switch.add_link(tgen.gears["r4"])


def setup_module(mod):
    "Sets up the pytest environment"

    logger.info("\n\n---- Starting CSPF tests ----\n")

    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for rname, router in router_list.items():
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_ISIS, os.path.join(CWD, "{}/isisd.conf".format(rname))
        )
        if rname == "r1":
            router.load_config(
                TopoRouter.RD_SHARP, os.path.join(CWD, "{}/sharpd.conf".format("r1"))
            )

    # Initialize all routers.
    tgen.start_router()


def teardown_module():
    "Teardown the pytest environment"

    tgen = get_topogen()
    tgen.stop_topology()

    logger.info("\n\n---- CSPF tests End ----\n")


def compare_ted_json_output(tgen, rname, fileref):
    "Compare TED JSON output"

    logger.info('Comparing router "%s" TED output', rname)

    filename = "{}/reference/{}".format(CWD, fileref)
    expected = json.loads(open(filename).read())
    command = "show sharp ted json"

    # Run test function until we get an result. Wait at most 60 seconds.
    test_func = partial(topotest.router_json_cmp, tgen.gears[rname], command, expected)
    _, diff = topotest.run_and_expect(test_func, None, count=60, wait=2)
    assertmsg = '"{}" TED JSON output mismatches the expected result'.format(rname)
    assert diff is None, assertmsg


def compare_cspf_output(tgen, rname, fileref, src, dst, cost, bw=""):
    "Compare CSPF output"

    logger.info('Comparing router "%s" CSPF output', rname)

    filename = "{}/reference/{}".format(CWD, fileref)
    expected = open(filename).read()
    command = "show sharp cspf source {} destination {} {} {}".format(
        src, dst, cost, bw
    )

    # Run test function until we get an result. Wait at most 60 seconds.
    test_func = partial(
        topotest.router_output_cmp, tgen.gears[rname], command, expected
    )
    result, diff = topotest.run_and_expect(test_func, "", count=5, wait=2)
    assert result, "CSPF output mismatches the expected result on {}:\n{}".format(
        rname, diff
    )


def setup_testcase(msg):
    "Setup test case"

    logger.info(msg)
    tgen = get_topogen()

    # Skip if previous fatal error condition is raised
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    return tgen


# Note that all routers must discover the same Network Topology, so the same TED.


def test_step1():
    "Step1: Check initial topology"

    tgen = setup_testcase("Step1: test initial IS-IS TE Data Base import")
    tgen.net["r1"].cmd('vtysh -c "sharp import-te"')

    compare_ted_json_output(tgen, "r1", "sharp-ted.json")


def test_step2():
    "Step2: Test CSPF from r1 to r4 for IPv4 with various metric"

    tgen = setup_testcase("Step2: CSPF(r1, r4, IPv4)")

    compare_cspf_output(
        tgen, "r1", "cspf-ipv4-metric.txt", "10.0.0.1", "10.0.255.4", "metric 50"
    )
    compare_cspf_output(
        tgen, "r1", "cspf-ipv4-te-metric.txt", "10.0.255.1", "10.0.4.4", "te-metric 50"
    )
    compare_cspf_output(
        tgen, "r1", "cspf-ipv4-delay.txt", "10.0.255.1", "10.0.255.4", "delay 50000"
    )
    compare_cspf_output(
        tgen,
        "r1",
        "cspf-ipv4-delay.txt",
        "10.0.255.1",
        "10.0.255.4",
        "delay 50000",
        "rsv 7 1000000",
    )


def test_step3():
    "Step3: Test CSPF from r1 to r4 for IPv6 with various metric"

    tgen = setup_testcase("Step2: CSPF(r1, r4, IPv6)")

    compare_cspf_output(
        tgen,
        "r1",
        "cspf-ipv6-metric.txt",
        "2001:db8:1::1:1",
        "2001:db8::4",
        "metric 50",
    )
    compare_cspf_output(
        tgen,
        "r1",
        "cspf-ipv6-te-metric.txt",
        "2001:db8::1",
        "2001:db8:5::3:4",
        "te-metric 80",
    )
    compare_cspf_output(
        tgen, "r1", "cspf-ipv6-delay.txt", "2001:db8::1", "2001:db8::4", "delay 80000"
    )
    compare_cspf_output(
        tgen,
        "r1",
        "cspf-ipv6-delay.txt",
        "2001:db8::1",
        "2001:db8::4",
        "delay 80000",
        "rsv 7 1000000",
    )


def test_step4():
    "Step4: Test CSPF from r1 to r4 with no possible path"

    tgen = setup_testcase("Step2: CSPF(r1, r4, failure)")

    compare_cspf_output(
        tgen, "r1", "cspf-failed.txt", "10.0.255.1", "10.0.255.4", "metric 10"
    )
    compare_cspf_output(
        tgen, "r1", "cspf-failed.txt", "2001:db8::1", "2001:db8::4", "te-metric 50"
    )
    compare_cspf_output(
        tgen, "r1", "cspf-failed.txt", "10.0.255.1", "10.0.255.4", "delay 5000"
    )
    compare_cspf_output(
        tgen,
        "r1",
        "cspf-failed.txt",
        "2001:db8::1",
        "2001:db8::4",
        "delay 80000",
        "rsv 7 10000000",
    )
    compare_cspf_output(
        tgen, "r1", "cspf-failed-src.txt", "10.0.0.3", "10.0.255.4", "metric 10"
    )
    compare_cspf_output(
        tgen, "r1", "cspf-failed-dst.txt", "10.0.0.1", "10.0.4.40", "metric 10"
    )
    compare_cspf_output(
        tgen, "r1", "cspf-failed-same.txt", "10.0.0.1", "10.0.0.1", "metric 10"
    )


def test_memory_leak():
    "Run the memory leak test and report results."

    tgen = get_topogen()
    if not tgen.is_memleak_enabled():
        pytest.skip("Memory leak test/report is disabled")

    tgen.report_memory_leaks()


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2020 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation,
# Inc. ("NetDEF") in this file.
#

"""
Following tests are covered to test multicast pim sm:

Test steps
- Create topology (setup module)
- Bring up topology

Following tests are covered:
1. TC_17: Verify (*,G) and (S,G) present and multicast traffic resume,
    after restart of PIMd daemon
2. TC_18: Verify (*,G) and (S,G) present and multicast traffic resume after
    FRR service stop and start
3. TC_10: Verify SPT switchover working when RPT and SPT path is
    different
4. TC_15: Verify (S,G) and (*,G) mroute after shut / no shut of upstream
    interfaces
5. TC_7: Verify mroute detail when receiver is present
    outside of FRR
6. TC_8: Verify mroute when FRR is acting as FHR and LHR
7. TC_20: Verify mroute detail when 5 different receiver joining
    same source
8. TC_22: Verify OIL and IIF detail updated in (S,G) mroute after shut
    and no shut of the source interface
"""

import os
import sys
import time
import pytest

pytestmark = [pytest.mark.pimd]

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# Required to instantiate the topology builder class.

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    step,
    reset_config_on_routers,
    shutdown_bringup_interface,
    kill_router_daemons,
    start_router,
    start_router_daemons,
    stop_router,
    required_linux_kernel_version,
)
from lib.pim import (
    create_pim_config,
    create_igmp_config,
    verify_igmp_groups,
    verify_mroutes,
    get_pim_interface_traffic,
    verify_upstream_iif,
    verify_pim_neighbors,
    verify_pim_state,
    clear_mroute,
    clear_pim_interface_traffic,
    McastTesterHelper,
)
from lib.topolog import logger
from lib.topojson import build_config_from_json


TOPOLOGY = """


    i4-----c1-------------c2---i5
            |              |
            |              |
    i1-----l1------r2-----f1---i2
       |    |      |       |
       |    |      |       |
      i7    i6     i3     i8

    Description:
    i1, i2, i3. i4, i5, i6, i7, i8 - FRR running iperf to send IGMP
                                     join and traffic
    l1 - LHR
    f1 - FHR
    r2 - FRR router
    c1 - FRR router
    c2 - FRR router
"""

# Global variables
GROUP_RANGE = "225.0.0.0/8"
IGMP_JOIN = "225.1.1.1"
GROUP_RANGE_1 = [
    "225.1.1.1/32",
    "225.1.1.2/32",
    "225.1.1.3/32",
    "225.1.1.4/32",
    "225.1.1.5/32",
]
IGMP_JOIN_RANGE_1 = ["225.1.1.1", "225.1.1.2", "225.1.1.3", "225.1.1.4", "225.1.1.5"]
GROUP_RANGE_2 = [
    "226.1.1.1/32",
    "226.1.1.2/32",
    "226.1.1.3/32",
    "226.1.1.4/32",
    "226.1.1.5/32",
]
IGMP_JOIN_RANGE_2 = ["226.1.1.1", "226.1.1.2", "226.1.1.3", "226.1.1.4", "226.1.1.5"]

GROUP_RANGE_3 = [
    "227.1.1.1/32",
    "227.1.1.2/32",
    "227.1.1.3/32",
    "227.1.1.4/32",
    "227.1.1.5/32",
]
IGMP_JOIN_RANGE_3 = ["227.1.1.1", "227.1.1.2", "227.1.1.3", "227.1.1.4", "227.1.1.5"]


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """

    # Required linux kernel version for this suite to run.
    result = required_linux_kernel_version("4.19")
    if result is not True:
        pytest.skip("Kernel version should be >= 4.19")

    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)
    logger.info("Master Topology: \n {}".format(TOPOLOGY))

    logger.info("Running setup_module to create topology")

    json_file = "{}/multicast_pim_sm_topo2.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # XXX Replace this using "with McastTesterHelper()... " in each test if possible.
    global app_helper
    app_helper = McastTesterHelper(tgen)

    logger.info("Running setup_module() done")


def teardown_module():
    """Teardown the pytest environment"""

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    app_helper.cleanup()

    # Stop toplogy and Remove tmp files
    tgen.stop_topology()

    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


#####################################################
#
#   Testcases
#
#####################################################


def verify_state_incremented(state_before, state_after):
    """
    API to compare interface traffic state incrementing

    Parameters
    ----------
    * `state_before` : State dictionary for any particular instance
    * `state_after` : State dictionary for any particular instance
    """

    for router, state_data in state_before.items():
        for state, _ in state_data.items():
            if state_before[router][state] >= state_after[router][state]:
                errormsg = (
                    "[DUT: %s]: state %s value has not"
                    " incremented, Initial value: %s, "
                    "Current value: %s [FAILED!!]"
                    % (
                        router,
                        state,
                        state_before[router][state],
                        state_after[router][state],
                    )
                )
                return errormsg

            logger.info(
                "[DUT: %s]: State %s value is "
                "incremented, Initial value: %s, Current value: %s"
                " [PASSED!!]",
                router,
                state,
                state_before[router][state],
                state_after[router][state],
            )

    return True


def test_verify_mroute_and_traffic_when_pimd_restarted_p2(request):
    """
    TC_17: Verify (*,G) and (S,G) present and multicast traffic resume,
    after restart of PIMd daemon
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for (226.1.1.1-5) in c1")
    step("Configure static RP for (232.1.1.1-5) in c2")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_2,
                    }
                ]
            }
        },
        "c2": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c2"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Enable IGMP on FRR1 interface and send IGMP join "
        "(226.1.1.1-5) and (232.1.1.1-5)"
    )
    step(
        "Configure IGMP interface on FRR3 and send IGMP join"
        " for group (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_dict = {
        "f1": {
            "igmp": {
                "interfaces": {
                    "f1-i8-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        }
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    input_join = {"i1": "i1-l1-eth0", "i8": "i8-f1-eth0"}

    for recvr, recvr_intf in input_join.items():
        result = app_helper.run_join(recvr, _IGMP_JOIN_RANGE, join_intf=recvr_intf)
        assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step(
        "Connect one source to c2 and send multicast traffic all"
        " the receivers (226.1.1.1-5, 232.1.1.1-5)"
    )
    step(
        "Send multicast traffic from FRR3 to all the receivers "
        "(226.1.1.1-5, 232.1.1.1-5)"
    )

    input_src = {"i2": "i2-f1-eth0", "i5": "i5-c2-eth0"}

    for src, src_intf in input_src.items():
        result = app_helper.run_traffic(src, _IGMP_JOIN_RANGE, bind_intf=src_intf)
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    # Verifying mroutes before PIMd restart, fetching uptime

    source = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"},
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-i8-eth2"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Restart Pimd process on FRR3 node")
    kill_router_daemons(tgen, "f1", ["pimd"])
    start_router_daemons(tgen, "f1", ["pimd"])

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "After restart of PIMd verify pim nbr is up , IGMP groups"
        " received , and (*,G) (S,G) entries populated again ,"
        " Verify using 'show ip pim neighbor' , 'show ip igmp groups'"
        " 'show ip mroute'"
    )

    result = verify_pim_neighbors(tgen, topo)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    dut = "f1"
    interface = "f1-i8-eth2"
    result = verify_igmp_groups(tgen, dut, interface, _IGMP_JOIN_RANGE)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Stop the traffic and restart PIMd immediately on FRR3 node")
    dut = "i2"
    intf = "i2-f1-eth0"
    shutdown_bringup_interface(tgen, dut, intf, False)

    kill_router_daemons(tgen, "f1", ["pimd"])
    start_router_daemons(tgen, "f1", ["pimd"])

    step(
        "After PIM process come , all the none of (S,G) mroute should"
        " present on FRR3 'show ip mroute' "
    )

    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    input_dict = [
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "none"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
            expected=False,
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: [{}]: mroute (S, G) should not be present in mroute table \n "
            "Found: {}".format(tc_name, data["dut"], result)
        )

    write_test_footer(tc_name)


def test_verify_mroute_and_traffic_when_frr_restarted_p2(request):
    """
    TC_18: Verify (*,G) and (S,G) present and multicast traffic resume after
    FRR service stop and start
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for (226.1.1.1-5) in c1")
    step("Configure static RP for (232.1.1.1-5) in c2")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_2,
                    }
                ]
            }
        },
        "c2": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c2"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Enable IGMP on FRR1 interface and send IGMP join "
        "(226.1.1.1-5) and (232.1.1.1-5)"
    )
    step(
        "Configure IGMP interface on FRR3 and send IGMP join"
        " for group (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_dict = {
        "f1": {
            "igmp": {
                "interfaces": {
                    "f1-i8-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        }
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    input_join = {"i1": "i1-l1-eth0", "i8": "i8-f1-eth0"}

    for recvr, recvr_intf in input_join.items():
        result = app_helper.run_join(recvr, _IGMP_JOIN_RANGE, join_intf=recvr_intf)
        assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step(
        "Connect one source to c2 and send multicast traffic all"
        " the receivers (226.1.1.1-5, 232.1.1.1-5)"
    )
    step(
        "Send multicast traffic from FRR3 to all the receivers "
        "(226.1.1.1-5, 232.1.1.1-5)"
    )

    input_src = {"i2": "i2-f1-eth0", "i5": "i5-c2-eth0"}

    for src, src_intf in input_src.items():
        result = app_helper.run_traffic(src, _IGMP_JOIN_RANGE, bind_intf=src_intf)
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Verifying mroutes before FRR restart, fetching uptime")

    source = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"},
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-i8-eth2"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Stop and Start the FRR services on FRR3 node")
    stop_router(tgen, "f1")
    start_router(tgen, "f1")

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "After stop and start of FRR service verify pim nbr is up "
        "IGMP groups received , and (*,G) (S,G) entries populated again"
        " Verify using 'show ip pim neighbor' , 'show ip igmp groups'"
        " 'show ip mroute'"
    )

    result = verify_pim_neighbors(tgen, topo)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    dut = "f1"
    interface = "f1-i8-eth2"
    result = verify_igmp_groups(tgen, dut, interface, _IGMP_JOIN_RANGE)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Stop the traffic and stop and start the FRR services on" " FRR3 node")
    shutdown_bringup_interface(tgen, "i2", "i2-f1-eth0", False)

    stop_router(tgen, "f1")
    start_router(tgen, "f1")

    step(
        "After stop and start of FRR services , all the none of (S,G)"
        " mroute should present on FRR3 node verify using "
        "'show ip mroute'"
    )

    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    input_dict = [
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "none"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
            expected=False,
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: [{}]: mroute (S, G) should not be present in mroute table \n "
            "Found: {}".format(tc_name, data["dut"], result)
        )

    write_test_footer(tc_name)


def test_verify_SPT_switchover_when_RPT_and_SPT_path_is_different_p0(request):
    """
    TC_10: Verify SPT switchover working when RPT and SPT path is
    different
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for (226.1.1.1-5) and " "(232.1.1.1-5) in c2")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c2": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c2"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": _GROUP_RANGE,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Enable IGMP on FRR1 interface and send IGMP join "
        "(226.1.1.1-5) and (232.1.1.1-5)"
    )

    result = app_helper.run_join("i1", _IGMP_JOIN_RANGE, "l1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("Send multicast traffic from FRR3 to '226.1.1.1-5'" ", '232.1.1.1-5' receiver")

    step("registerRx and registerStopTx value before traffic sent")
    state_dict = {"c2": {"c2-f1-eth1": ["registerRx", "registerStopTx"]}}
    state_before = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_before, dict
    ), "Testcase {} : Failed \n state_before is not dictionary \nError: {}".format(
        tc_name, result
    )

    result = app_helper.run_traffic("i2", _IGMP_JOIN_RANGE, "f1")
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Verify in FRR3 sending initial packet to RP using"
        " 'show ip mroute' and mroute OIL is towards RP."
    )

    result = verify_mroutes(
        tgen,
        "f1",
        "10.0.5.2",
        _IGMP_JOIN_RANGE,
        "f1-i2-eth1",
        ["f1-c2-eth0", "f1-r2-eth3"],
    )
    assert result is True, "Testcase {} : " "Failed Error: {}".format(tc_name, result)

    result = verify_mroutes(
        tgen, "f1", "10.0.5.2", _IGMP_JOIN_RANGE, "f1-i2-eth1", "f1-r2-eth3"
    )
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        " After spt switchover traffic is flowing between"
        " (LHR(FRR1)-FHR(FRR3)) and (S,G) OIL is updated toward FRR1"
        " 'show ip mroute' and 'show ip pim upstream'"
    )

    source = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict = [
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-r2-eth3"},
        {"dut": "l1", "src_address": source, "iif": "l1-r2-eth4", "oil": "l1-i1-eth1"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Stop the traffic to all the receivers")

    app_helper.stop_host("i2")

    step(
        "Null register packet being send periodically from FRR3 to RP, "
        "verify using show ip mroute on RP, have (S, G) entries null OIL"
        " 'show ip mroute' and verify show ip pim interface traffic"
        "(In RP Register msg should be received and Register stop should"
        " be transmitted)"
    )
    input_dict = [
        {"dut": "c2", "src_address": source, "iif": "c2-f1-eth1", "oil": "none"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("registerRx and registerStopTx value after traffic sent")
    state_after = get_pim_interface_traffic(tgen, state_dict)
    assert isinstance(
        state_after, dict
    ), "Testcase {} : Failed \n state_before is not dictionary \nError: {}".format(
        tc_name, result
    )

    result = verify_state_incremented(state_before, state_after)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_verify_mroute_after_shut_noshut_of_upstream_interface_p1(request):
    """
    TC_15: Verify (S,G) and (*,G) mroute after shut / no shut of upstream
    interfaces
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for (226.1.1.1-5) in c1")
    step("Configure static RP for (232.1.1.1-5) in c2")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_2,
                    }
                ]
            }
        },
        "c2": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c2"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Enable IGMP on FRR1 interface and send IGMP join "
        "(226.1.1.1-5) and (232.1.1.1-5)"
    )
    step(
        "Configure IGMP interface on FRR3 and send IGMP join"
        " for group (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_dict = {
        "f1": {
            "igmp": {
                "interfaces": {
                    "f1-i8-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        }
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    input_join = {"i1": "i1-l1-eth0", "i8": "i8-f1-eth0"}

    for recvr, recvr_intf in input_join.items():
        result = app_helper.run_join(recvr, _IGMP_JOIN_RANGE, join_intf=recvr_intf)
        assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step(
        "Connect one source to c2 and send multicast traffic all"
        " the receivers (226.1.1.1-5, 232.1.1.1-5)"
    )
    step(
        "Send multicast traffic from FRR3 to all the receivers "
        "(226.1.1.1-5, 232.1.1.1-5)"
    )

    input_src = {"i2": "i2-f1-eth0", "i5": "i5-c2-eth0"}

    for src, src_intf in input_src.items():
        result = app_helper.run_traffic(src, _IGMP_JOIN_RANGE, bind_intf=src_intf)
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "FRR3 (S,G) has one OIL for local receiver one toward c2"
        " verify 'show ip mroute' and 'show ip pim upstream'"
    )

    source = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"},
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-i8-eth2"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], IGMP_JOIN_RANGE_2
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Shut and No shut interface connected from FHR (FRR3)" " to c2")
    dut = "f1"
    intf = "f1-c2-eth0"
    shutdown_bringup_interface(tgen, dut, intf, False)
    shutdown_bringup_interface(tgen, dut, intf, True)

    step("Shut and No shut interface connected from LHR (FRR1)" " to c1")
    dut = "l1"
    intf = "l1-c1-eth0"
    shutdown_bringup_interface(tgen, dut, intf, False)
    shutdown_bringup_interface(tgen, dut, intf, True)

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Shut and No shut FRR1 and FRR3 interface")
    shutdown_bringup_interface(tgen, "l1", "l1-r2-eth4", False)
    shutdown_bringup_interface(tgen, dut, intf, True)

    shutdown_bringup_interface(tgen, "f1", "f1-r2-eth3", False)
    shutdown_bringup_interface(tgen, dut, intf, True)

    step(
        "After shut/no shut of interface , verify traffic resume to all"
        "the receivers (S,G) OIL update for all the receivers"
    )

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Shut FRR1, FRR3 interface , clear mroute in FRR1"
        " and No shut FRR1, FRR3 interface "
    )
    dut = "l1"
    intf = "l1-r2-eth4"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "f1"
    intf = "f1-r2-eth3"
    shutdown_bringup_interface(tgen, dut, intf, False)

    dut = "l1"
    intf = "l1-r2-eth4"
    shutdown_bringup_interface(tgen, dut, intf, True)

    dut = "f1"
    intf = "f1-r2-eth3"
    shutdown_bringup_interface(tgen, dut, intf, True)

    clear_mroute(tgen, "l1")
    clear_mroute(tgen, "l1")

    step(
        "After no shut, verify traffic resume to all the receivers"
        " (S,G) OIL update for all the receivers"
    )

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Shut and no shut upstream interface from FRR1 to FRR2 and "
        "cisco immediate after mroute/upstream got cleared"
    )

    dut = "l1"
    intf_l1_r2 = "l1-r2-eth4"
    shutdown_bringup_interface(tgen, dut, intf_l1_r2, False)

    intf_l1_c1 = "l1-c1-eth0"
    shutdown_bringup_interface(tgen, dut, intf_l1_c1, False)

    result = verify_upstream_iif(
        tgen, "l1", "Unknown", source, IGMP_JOIN_RANGE_2, expected=False
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: Upstream IIF should be unknown \n "
        "Found: {}".format(tc_name, "l1", result)
    )

    step("No shut the Source interface just after the upstream is expired" " from FRR1")
    shutdown_bringup_interface(tgen, dut, intf_l1_r2, True)
    shutdown_bringup_interface(tgen, dut, intf_l1_c1, True)

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Stop the traffic to all the receivers")
    app_helper.stop_all_hosts()

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
            expected=False,
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: [{}]: mroute (S, G) should not be present in mroute table \n "
            "Found: {}".format(tc_name, data["dut"], result)
        )

    write_test_footer(tc_name)


def test_verify_mroute_when_receiver_is_outside_frr_p0(request):
    """
    TC_7: Verify mroute detail when receiver is present
    outside of FRR
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP on c1 for group range " "(226.1.1.1-5) and (232.1.1.1-5)")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": _GROUP_RANGE,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Enable IGMP on FRR1 interface and send IGMP join"
        " (226.1.1.1-5) and (232.1.1.1-5)"
    )
    result = app_helper.run_join("i1", _IGMP_JOIN_RANGE, "l1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step(
        "Send multicast traffic from FRR3 to all the receivers "
        "(226.1.1.1-5) and (232.1.1.1-5)"
    )
    result = app_helper.run_traffic("i2", _IGMP_JOIN_RANGE, "f1")
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Configure one more receiver in c2 enable IGMP and send"
        " join (226.1.1.1-5) and (232.1.1.1-5)"
    )
    input_dict = {
        "c2": {
            "igmp": {
                "interfaces": {
                    "c2-i5-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        }
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    result = app_helper.run_join("i5", _IGMP_JOIN_RANGE, "c2")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("FRR1 has 10 (*.G) and 10 (S,G) verify using 'show ip mroute count'")
    step(
        "All the receiver are receiving traffic on FRR1 and (S,G) OIL is toward"
        "receivers, verify using 'show ip mroute' 'show ip pim upstream'"
    )
    step(
        "All the receiver are receiving traffic on c2 and (S,G) OIL is "
        "toward receivers, verify using 'show ip mroute' 'show ip pim upstream'"
    )

    source = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict = [
        {"dut": "l1", "src_address": "*", "iif": "l1-c1-eth0", "oil": "l1-i1-eth1"},
        {"dut": "l1", "src_address": source, "iif": "l1-r2-eth4", "oil": "l1-i1-eth1"},
        {"dut": "c2", "src_address": "*", "iif": "c2-c1-eth0", "oil": "c2-i5-eth2"},
        {"dut": "c2", "src_address": source, "iif": "c2-f1-eth1", "oil": "c2-i5-eth2"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "FRR3 has (S,G) OIL created toward c1/c2 receiver and FRR1 receiver"
        "'show ip pim state'"
    )
    input_dict = [
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-c2-eth0"},
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-r2-eth3"},
    ]
    for data in input_dict:
        result = verify_pim_state(
            tgen,
            data["dut"],
            data["iif"],
            data["oil"],
            _IGMP_JOIN_RANGE,
            data["src_address"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_verify_mroute_when_FRR_is_FHR_and_LHR_p0(request):
    """
    TC_8: Verify mroute when FRR is acting as FHR and LHR
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for group range (226.1.1.1-5) and " "(232.1.1.1-5) on c1")
    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": _GROUP_RANGE,
                    }
                ]
            }
        }
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Enable IGMP on FRR1 interface and send IGMP join (226.1.1.1-5)"
        " and (232.1.1.1-5)"
    )
    step(
        "Configure receiver on FRR3 with igmp and pim enabled and "
        "send IGMP join (226.1.1.1-5) and (232.1.1.1-5)"
    )
    step(
        "Send multicast traffic from FRR3 to all the receivers "
        "(226.1.1.1-5) and (232.1.1.1-5)"
    )

    step("Send IGMP join (226.1.1.1-5, 232.1.1.1-5) to LHR(l1)")
    result = app_helper.run_join("i1", _IGMP_JOIN_RANGE, "l1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("Send multicast traffic from FRR3 to '226.1.1.1-5'" ", '232.1.1.1-5' receiver")
    result = app_helper.run_traffic("i2", _IGMP_JOIN_RANGE, "f1")
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Configure receiver in f1 enable IGMP and send"
        " join (226.1.1.1-5) and (232.1.1.1-5)"
    )

    step("Configure one IGMP interface on f1 node and send IGMP" " join (225.1.1.1)")
    input_dict = {
        "f1": {
            "igmp": {
                "interfaces": {
                    "f1-i8-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        }
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    result = app_helper.run_join("i8", _IGMP_JOIN_RANGE, "f1")
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)
    step(
        "l1 and f1 has 10 IGMP groups (226.1.1.1-5, 232.1.1.1-5),"
        " verify using 'show ip igmp groups'"
    )
    dut = "l1"
    interface = "l1-i1-eth1"
    result = verify_igmp_groups(tgen, dut, interface, _IGMP_JOIN_RANGE)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    dut = "f1"
    interface = "f1-i8-eth2"
    result = verify_igmp_groups(tgen, dut, interface, _IGMP_JOIN_RANGE)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "l1 , f1 has 10 (*,G) and 10 (S,G) for groups "
        "(226.1.1.1-5, 232.1.1.1-5), verify using "
        " 'show ip mroute'"
    )

    source = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"},
        {"dut": "f1", "src_address": source, "iif": "f1-i2-eth1", "oil": "f1-r2-eth3"},
        {"dut": "l1", "src_address": "*", "iif": "l1-c1-eth0", "oil": "l1-i1-eth1"},
        {"dut": "l1", "src_address": source, "iif": "l1-r2-eth4", "oil": "l1-i1-eth1"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Join timer is running in FHR and LHR , verify using" " 'show ip pim state'")

    for data in input_dict:
        result = verify_pim_state(
            tgen,
            data["dut"],
            data["iif"],
            data["oil"],
            _IGMP_JOIN_RANGE,
            data["src_address"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    # Stop the multicast traffic
    step("Stop the traffic to all the receivers")
    app_helper.stop_all_hosts()

    step(
        "After traffic stopped , verify (*,G) entries are not flushed"
        " out from FRR1 node verify using 'show ip mroute' "
    )

    input_dict = [
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"},
        {"dut": "l1", "src_address": "*", "iif": "l1-c1-eth0", "oil": "l1-i1-eth1"},
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert (
            result is True
        ), "Testcase {} : Failed Error mroutes were flushed.".format(tc_name)

    step(
        "After traffic stopped , verify (S,G) entries are flushed out"
        " from FRR1 node verify using 'show ip mroute' "
    )

    input_dict = [
        {"dut": "l1", "src_address": source, "iif": "l1-r2-eth4", "oil": "l1-i1-eth1"},
        {"dut": "f1", "src_address": source, "iif": "i2-f1-eth0", "oil": "f1-r2-eth3"},
    ]

    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
            expected=False,
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: [{}]: mroute (S, G) should not be present in mroute table \n "
            "Found: {}".format(tc_name, data["dut"], result)
        )

    write_test_footer(tc_name)


def test_verify_mroute_when_5_different_receiver_joining_same_sources_p0(request):
    """
    TC_20: Verify mroute detail when 5 different receiver joining
    same source
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for (226.1.1.1-5) in c1")
    step("Configure static RP for (232.1.1.1-5) in c2")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_2,
                    }
                ]
            }
        },
        "c2": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c2"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Configure 2 IGMP interface on FRR1 and send IGMP join"
        "for group (226.1.1.1-5, 232.1.1.1-5) from both the interface"
    )
    step(
        "Configure 2 IGMP interface on FRR3 and send IGMP join for"
        " group (226.1.1.1-5, 232.1.1.1-5) from both the interface"
    )
    step(
        "Configure 1 IGMP interface on c2 and send IGMP join for"
        "group (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_dict = {
        "f1": {
            "igmp": {
                "interfaces": {
                    "f1-i8-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    },
                    "f1-i2-eth1": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    },
                }
            }
        },
        "l1": {
            "igmp": {
                "interfaces": {
                    "l1-i6-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        },
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    input_join = {
        "i1": "i1-l1-eth0",
        "i6": "i6-l1-eth0",
        "i8": "i8-f1-eth0",
        "i2": "i2-f1-eth0",
    }

    for recvr, recvr_intf in input_join.items():
        result = app_helper.run_join(recvr, _IGMP_JOIN_RANGE, join_intf=recvr_intf)
        assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("Configure one source in FRR2 , one in c1")
    step(
        "Send multicast traffic from both the sources to all the"
        "receivers (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_src = {"i3": "i3-r2-eth0"}

    for src, src_intf in input_src.items():
        result = app_helper.run_traffic(src, _IGMP_JOIN_RANGE, bind_intf=src_intf)
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)
    step(
        "After all the IGMP groups received with correct port using"
        " 'show ip igmp groups' in FRR1, FRR3, c2"
    )
    dut = "l1"
    interface = "l1-i6-eth2"
    result = verify_igmp_groups(tgen, dut, interface, _IGMP_JOIN_RANGE)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    dut = "f1"
    interface = "f1-i8-eth2"
    result = verify_igmp_groups(tgen, dut, interface, _IGMP_JOIN_RANGE)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "(*,G) entries got created with upstream interface RP connected"
        " port using 'show ip pim upstream' in FRR1, FRR3, c2"
    )
    step(
        "(S,G) entries created for all the receiver after starting the"
        " source , traffic is reaching to all the receiver , verify OIL"
        " of (S,G) is receiver port using 'show ip mroute' in FRR1, "
        "FRR3 c2"
    )

    source = topo["routers"]["i3"]["links"]["r2"]["ipv4"].split("/")[0]
    input_dict_all = [
        {
            "dut": "l1",
            "src_address": source,
            "iif": ["l1-r2-eth4", "l1-c1-eth0"],
            "oil": ["l1-i1-eth1", "l1-i6-eth2"],
        },
        {"dut": "f1", "src_address": source, "iif": "f1-r2-eth3", "oil": "f1-i8-eth2"},
    ]
    for data in input_dict_all:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict_all:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Shut the receiver interface one by one on FRR1 node")
    shutdown_bringup_interface(tgen, "l1", "l1-i1-eth1", False)
    shutdown_bringup_interface(tgen, "l1", "l1-i6-eth2", False)

    step(
        "After shut the receiver port verify traffic is stopped immediately"
        " and (S,G) got timeout immediately in FRR1, FRR3, c2"
    )
    input_dict = [
        {"dut": "l1", "src_address": source, "iif": "l1-r2-eth4", "oil": "l1-i1-eth1"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            IGMP_JOIN_RANGE_2,
            data["iif"],
            data["oil"],
            expected=False,
        )
        assert result is not True, (
            "Testcase {} : Failed \n "
            "Expected: [{}]: mroute (S, G) should not be present in mroute table \n "
            "Found: {}".format(tc_name, data["dut"], result)
        )

    step(
        "No traffic impact observed on other receivers verify using"
        " 'show ip mroute' "
    )
    input_dict = [
        {"dut": "f1", "src_address": source, "iif": "f1-r2-eth3", "oil": "f1-i8-eth2"}
    ]
    for data in input_dict:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("No shut the receiver interface one by one on FRR1 node")
    shutdown_bringup_interface(tgen, "l1", "l1-i1-eth1", True)
    shutdown_bringup_interface(tgen, "l1", "l1-i6-eth2", True)

    step(
        "After no shut of receivers all the mroute entries got populated"
        ", no duplicate entries present in mroute"
    )

    for data in input_dict_all:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


def test_verify_oil_iif_for_mroute_after_shut_noshut_source_interface_p1(request):
    """
    TC_22: Verify OIL and IIF detail updated in (S,G) mroute after shut
    and no shut of the source interface
    """

    tgen = get_topogen()
    tc_name = request.node.name
    write_test_header(tc_name)

    # Don"t run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    # Creating configuration from JSON
    app_helper.stop_all_hosts()
    clear_mroute(tgen)
    reset_config_on_routers(tgen)
    clear_pim_interface_traffic(tgen, topo)

    step("Configure static RP for (226.1.1.1-5) in c1")
    step("Configure static RP for (232.1.1.1-5) in c2")

    _GROUP_RANGE = GROUP_RANGE_2 + GROUP_RANGE_3
    _IGMP_JOIN_RANGE = IGMP_JOIN_RANGE_2 + IGMP_JOIN_RANGE_3

    input_dict = {
        "c1": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c1"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_2,
                    }
                ]
            }
        },
        "c2": {
            "pim": {
                "rp": [
                    {
                        "rp_addr": topo["routers"]["c2"]["links"]["lo"]["ipv4"].split(
                            "/"
                        )[0],
                        "group_addr_range": GROUP_RANGE_3,
                    }
                ]
            }
        },
    }

    result = create_pim_config(tgen, topo, input_dict)
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "Configure IGMP interface on FRR1 and FRR3 and send IGMP join"
        " for group (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_dict = {
        "f1": {
            "igmp": {
                "interfaces": {
                    "f1-i8-eth2": {
                        "igmp": {"version": "2", "query": {"query-interval": 15}}
                    }
                }
            }
        }
    }
    result = create_igmp_config(tgen, topo, input_dict)
    assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    input_join = {"i1": "i1-l1-eth0", "i8": "i8-f1-eth0"}

    for recvr, recvr_intf in input_join.items():
        result = app_helper.run_join(recvr, _IGMP_JOIN_RANGE, join_intf=recvr_intf)
        assert result is True, "Testcase {}: Failed Error: {}".format(tc_name, result)

    step("Configure 1 source in FRR1 , 1 in FRR3")
    step(
        "Send multicast traffic from both the sources to all the "
        "receivers (226.1.1.1-5, 232.1.1.1-5)"
    )

    input_src = {"i6": "i6-l1-eth0", "i2": "i2-f1-eth0"}

    for src, src_intf in input_src.items():
        result = app_helper.run_traffic(src, _IGMP_JOIN_RANGE, bind_intf=src_intf)
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step(
        "*,G) is created and (S,G) created on FRR1 and FRR3 for both"
        " the source verify using 'show ip mroute' and "
        " 'show ip pim upstream' to check the upstream interface"
        " details"
    )

    source_i6 = topo["routers"]["i6"]["links"]["l1"]["ipv4"].split("/")[0]
    source_i2 = topo["routers"]["i2"]["links"]["f1"]["ipv4"].split("/")[0]
    input_dict_all = [
        {"dut": "l1", "src_address": "*", "iif": "l1-c1-eth0", "oil": "l1-i1-eth1"},
        {
            "dut": "l1",
            "src_address": source_i2,
            "iif": "l1-r2-eth4",
            "oil": "l1-i1-eth1",
        },
        {
            "dut": "l1",
            "src_address": source_i6,
            "iif": "l1-i6-eth2",
            "oil": "l1-i1-eth1",
        },
        {"dut": "f1", "src_address": "*", "iif": "f1-c2-eth0", "oil": "f1-i8-eth2"},
        {
            "dut": "f1",
            "src_address": source_i2,
            "iif": "f1-i2-eth1",
            "oil": "f1-i8-eth2",
        },
        {
            "dut": "f1",
            "src_address": source_i6,
            "iif": "f1-r2-eth3",
            "oil": "f1-i8-eth2",
        },
    ]
    for data in input_dict_all:
        result = verify_mroutes(
            tgen,
            data["dut"],
            data["src_address"],
            _IGMP_JOIN_RANGE,
            data["iif"],
            data["oil"],
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    for data in input_dict_all:
        result = verify_upstream_iif(
            tgen, data["dut"], data["iif"], data["src_address"], _IGMP_JOIN_RANGE
        )
        assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    step("Shut the source interface one by one on FRR1")
    shutdown_bringup_interface(tgen, "f1", "f1-i2-eth1", False)

    step(
        "After shut of ource interface from FRR3 verify all the (S,G) "
        "entries flushed out from FRR3 node 'show ip pim upstream' "
        " 'show ip mroute' "
    )

    result = verify_mroutes(
        tgen,
        "f1",
        source_i2,
        _IGMP_JOIN_RANGE,
        "f1-i2-eth1",
        "f1-i8-eth2",
        expected=False,
    )
    assert result is not True, (
        "Testcase {} : Failed \n "
        "Expected: [{}]: mroute (S, G) should not be present in mroute table \n "
        "Found: {}".format(tc_name, data["dut"], result)
    )

    result = verify_upstream_iif(
        tgen, "f1", "Unknown", "10.0.5.2", _IGMP_JOIN_RANGE, joinState="NotJoined"
    )
    assert result is True, "Testcase {} : Failed Error: {}".format(tc_name, result)

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

#
# Part of NetDEF Topology Tests
#
# Copyright (c) 2017 by
# Network Device Education Foundation, Inc. ("NetDEF")
#

r"""
customize.py: Simple FRR MPLS L3VPN test topology

                  |
             +----+----+
             |   ce1   |
             | 99.0.0.1|                              CE Router
             +----+----+
       192.168.1. | .2  ce1-eth0
                  | .1  r1-eth4
             +---------+
             |    r1   |
             | 1.1.1.1 |                              PE Router
             +----+----+
                  | .1  r1-eth0
                  |
            ~~~~~~~~~~~~~
          ~~     sw0     ~~
          ~~ 10.0.1.0/24 ~~
            ~~~~~~~~~~~~~
                  |10.0.1.0/24
                  |
                  | .2  r2-eth0
             +----+----+
             |    r2   |
             | 2.2.2.2 |                              P router
             +--+---+--+
    r2-eth2  .2 |   | .2  r2-eth1
         ______/     \______
        /                   \
  ~~~~~~~~~~~~~        ~~~~~~~~~~~~~
~~     sw2     ~~    ~~     sw1     ~~
~~ 10.0.3.0/24 ~~    ~~ 10.0.2.0/24 ~~
  ~~~~~~~~~~~~~        ~~~~~~~~~~~~~
        |                 /    |
         \      _________/     |
          \    /                \
r3-eth1 .3 |  | .3  r3-eth0      | .4 r4-eth0
      +----+--+---+         +----+----+
      |     r3    |         |    r4   |
      |  3.3.3.3  |         | 4.4.4.4 |               PE Routers
      +-----------+         +---------+
 192.168.1. | .1     192.168.1.  | .1    rX-eth4
            | .2                 | .2    ceX-eth0
      +-----+-----+         +----+-----+
      |    ce2    |         |   ce3    |
      | 99.0.0.2  |         | 99.0.0.3 |              CE Routers
      +-----+-----+         +----+-----+
            |                    |

"""

import os

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import get_topogen
from lib.topolog import logger
from lib.ltemplate import ltemplateRtrCmd

# Required to instantiate the topology builder class.


CWD = os.path.dirname(os.path.realpath(__file__))
# test name based on directory
TEST = os.path.basename(CWD)


def build_topo(tgen):
    "Build function"

    # This function only purpose is to define allocation and relationship
    # between routers, switches and hosts.
    #
    # Create P/PE routers
    tgen.add_router("r1")
    # check for mpls
    if tgen.hasmpls != True:
        logger.info("MPLS not available, tests will be skipped")
        return
    for routern in range(2, 5):
        tgen.add_router("r{}".format(routern))
    # Create CE routers
    for routern in range(1, 4):
        tgen.add_router("ce{}".format(routern))

    # CE/PE links
    tgen.add_link(tgen.gears["ce1"], tgen.gears["r1"], "ce1-eth0", "r1-eth4")
    tgen.add_link(tgen.gears["ce2"], tgen.gears["r3"], "ce2-eth0", "r3-eth4")
    tgen.add_link(tgen.gears["ce3"], tgen.gears["r4"], "ce3-eth0", "r4-eth4")

    # Create a switch with just one router connected to it to simulate a
    # empty network.
    switch = {}
    switch[0] = tgen.add_switch("sw0")
    switch[0].add_link(tgen.gears["r1"], nodeif="r1-eth0")
    switch[0].add_link(tgen.gears["r2"], nodeif="r2-eth0")

    switch[1] = tgen.add_switch("sw1")
    switch[1].add_link(tgen.gears["r2"], nodeif="r2-eth1")
    switch[1].add_link(tgen.gears["r3"], nodeif="r3-eth0")
    switch[1].add_link(tgen.gears["r4"], nodeif="r4-eth0")

    switch[1] = tgen.add_switch("sw2")
    switch[1].add_link(tgen.gears["r2"], nodeif="r2-eth2")
    switch[1].add_link(tgen.gears["r3"], nodeif="r3-eth1")


def ltemplatePreRouterStartHook():
    cc = ltemplateRtrCmd()
    tgen = get_topogen()
    logger.info("pre router-start hook")
    # check for mpls
    if tgen.hasmpls != True:
        logger.info("MPLS not available, skipping setup")
        return False
    logger.info("setup mpls input")
    return True


def ltemplatePostRouterStartHook():
    logger.info("post router-start hook")
    return True



import frrtest


class TestFlag(frrtest.TestMultiOut):
    program = "./test_peer_attr"


# List of tests can be generated by executing:
# $> ./test_peer_attr 2>&1 | sed -n 's/\\/\\\\/g; s/\S\+ \[test\] \(.\+\)/TestFlag.okfail(\x27\1\x27)/pg'
#
TestFlag.okfail("peer\\advertisement-interval")
TestFlag.okfail("peer\\capability dynamic")
TestFlag.okfail("peer\\capability extended-nexthop")
# TestFlag.okfail('peer\\capability extended-nexthop')
TestFlag.okfail("peer\\description")
TestFlag.okfail("peer\\disable-connected-check")
TestFlag.okfail("peer\\dont-capability-negotiate")
TestFlag.okfail("peer\\capability fqdn")
TestFlag.okfail("peer\\local-as")
TestFlag.okfail("peer\\local-as 1 no-prepend")
TestFlag.okfail("peer\\local-as 1 no-prepend replace-as")
TestFlag.okfail("peer\\override-capability")
TestFlag.okfail("peer\\passive")
TestFlag.okfail("peer\\password")
TestFlag.okfail("peer\\shutdown")
TestFlag.okfail("peer\\strict-capability-match")
TestFlag.okfail("peer\\timers")
TestFlag.okfail("peer\\timers connect")
TestFlag.okfail("peer\\update-source")
TestFlag.okfail("peer\\update-source")
TestFlag.okfail("peer\\ipv4-unicast\\addpath")
TestFlag.okfail("peer\\ipv4-multicast\\addpath")
TestFlag.okfail("peer\\ipv6-unicast\\addpath")
TestFlag.okfail("peer\\ipv6-multicast\\addpath")
TestFlag.okfail("peer\\ipv4-unicast\\allowas-in")
TestFlag.okfail("peer\\ipv4-multicast\\allowas-in")
TestFlag.okfail("peer\\ipv6-unicast\\allowas-in")
TestFlag.okfail("peer\\ipv6-multicast\\allowas-in")
TestFlag.okfail("peer\\ipv4-unicast\\allowas-in origin")
TestFlag.okfail("peer\\ipv4-multicast\\allowas-in origin")
TestFlag.okfail("peer\\ipv6-unicast\\allowas-in origin")
TestFlag.okfail("peer\\ipv6-multicast\\allowas-in origin")
TestFlag.okfail("peer\\ipv4-unicast\\as-override")
TestFlag.okfail("peer\\ipv4-multicast\\as-override")
TestFlag.okfail("peer\\ipv6-unicast\\as-override")
TestFlag.okfail("peer\\ipv6-multicast\\as-override")
TestFlag.okfail("peer\\ipv4-unicast\\attribute-unchanged as-path")
TestFlag.okfail("peer\\ipv4-multicast\\attribute-unchanged as-path")
TestFlag.okfail("peer\\ipv6-unicast\\attribute-unchanged as-path")
TestFlag.okfail("peer\\ipv6-multicast\\attribute-unchanged as-path")
TestFlag.okfail("peer\\ipv4-unicast\\attribute-unchanged next-hop")
TestFlag.okfail("peer\\ipv4-multicast\\attribute-unchanged next-hop")
TestFlag.okfail("peer\\ipv6-unicast\\attribute-unchanged next-hop")
TestFlag.okfail("peer\\ipv6-multicast\\attribute-unchanged next-hop")
TestFlag.okfail("peer\\ipv4-unicast\\attribute-unchanged med")
TestFlag.okfail("peer\\ipv4-multicast\\attribute-unchanged med")
TestFlag.okfail("peer\\ipv6-unicast\\attribute-unchanged med")
TestFlag.okfail("peer\\ipv6-multicast\\attribute-unchanged med")
TestFlag.okfail("peer\\ipv4-unicast\\attribute-unchanged as-path next-hop")
TestFlag.okfail("peer\\ipv4-multicast\\attribute-unchanged as-path next-hop")
TestFlag.okfail("peer\\ipv6-unicast\\attribute-unchanged as-path next-hop")
TestFlag.okfail("peer\\ipv6-multicast\\attribute-unchanged as-path next-hop")
TestFlag.okfail("peer\\ipv4-unicast\\attribute-unchanged as-path med")
TestFlag.okfail("peer\\ipv4-multicast\\attribute-unchanged as-path med")
TestFlag.okfail("peer\\ipv6-unicast\\attribute-unchanged as-path med")
TestFlag.okfail("peer\\ipv6-multicast\\attribute-unchanged as-path med")
TestFlag.okfail("peer\\ipv4-unicast\\attribute-unchanged as-path next-hop med")
TestFlag.okfail("peer\\ipv4-multicast\\attribute-unchanged as-path next-hop med")
TestFlag.okfail("peer\\ipv6-unicast\\attribute-unchanged as-path next-hop med")
TestFlag.okfail("peer\\ipv6-multicast\\attribute-unchanged as-path next-hop med")
TestFlag.okfail("peer\\ipv4-unicast\\capability orf prefix-list send")
TestFlag.okfail("peer\\ipv4-multicast\\capability orf prefix-list send")
TestFlag.okfail("peer\\ipv6-unicast\\capability orf prefix-list send")
TestFlag.okfail("peer\\ipv6-multicast\\capability orf prefix-list send")
TestFlag.okfail("peer\\ipv4-unicast\\capability orf prefix-list receive")
TestFlag.okfail("peer\\ipv4-multicast\\capability orf prefix-list receive")
TestFlag.okfail("peer\\ipv6-unicast\\capability orf prefix-list receive")
TestFlag.okfail("peer\\ipv6-multicast\\capability orf prefix-list receive")
TestFlag.okfail("peer\\ipv4-unicast\\capability orf prefix-list both")
TestFlag.okfail("peer\\ipv4-multicast\\capability orf prefix-list both")
TestFlag.okfail("peer\\ipv6-unicast\\capability orf prefix-list both")
TestFlag.okfail("peer\\ipv6-multicast\\capability orf prefix-list both")
TestFlag.okfail("peer\\ipv4-unicast\\default-originate")
TestFlag.okfail("peer\\ipv4-multicast\\default-originate")
TestFlag.okfail("peer\\ipv6-unicast\\default-originate")
TestFlag.okfail("peer\\ipv6-multicast\\default-originate")
TestFlag.okfail("peer\\ipv4-unicast\\default-originate route-map")
TestFlag.okfail("peer\\ipv4-multicast\\default-originate route-map")
TestFlag.okfail("peer\\ipv6-unicast\\default-originate route-map")
TestFlag.okfail("peer\\ipv6-multicast\\default-originate route-map")
TestFlag.okfail("peer\\ipv4-unicast\\distribute-list")
TestFlag.okfail("peer\\ipv4-multicast\\distribute-list")
TestFlag.okfail("peer\\ipv6-unicast\\distribute-list")
TestFlag.okfail("peer\\ipv6-multicast\\distribute-list")
TestFlag.okfail("peer\\ipv4-unicast\\distribute-list")
TestFlag.okfail("peer\\ipv4-multicast\\distribute-list")
TestFlag.okfail("peer\\ipv6-unicast\\distribute-list")
TestFlag.okfail("peer\\ipv6-multicast\\distribute-list")
TestFlag.okfail("peer\\ipv4-unicast\\filter-list")
TestFlag.okfail("peer\\ipv4-multicast\\filter-list")
TestFlag.okfail("peer\\ipv6-unicast\\filter-list")
TestFlag.okfail("peer\\ipv6-multicast\\filter-list")
TestFlag.okfail("peer\\ipv4-unicast\\filter-list")
TestFlag.okfail("peer\\ipv4-multicast\\filter-list")
TestFlag.okfail("peer\\ipv6-unicast\\filter-list")
TestFlag.okfail("peer\\ipv6-multicast\\filter-list")
TestFlag.okfail("peer\\ipv4-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-unicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv6-multicast\\maximum-prefix")
TestFlag.okfail("peer\\ipv4-unicast\\next-hop-self")
TestFlag.okfail("peer\\ipv4-multicast\\next-hop-self")
TestFlag.okfail("peer\\ipv6-unicast\\next-hop-self")
TestFlag.okfail("peer\\ipv6-multicast\\next-hop-self")
TestFlag.okfail("peer\\ipv4-unicast\\next-hop-self force")
TestFlag.okfail("peer\\ipv4-multicast\\next-hop-self force")
TestFlag.okfail("peer\\ipv6-unicast\\next-hop-self force")
TestFlag.okfail("peer\\ipv6-multicast\\next-hop-self force")
TestFlag.okfail("peer\\ipv4-unicast\\prefix-list")
TestFlag.okfail("peer\\ipv4-multicast\\prefix-list")
TestFlag.okfail("peer\\ipv6-unicast\\prefix-list")
TestFlag.okfail("peer\\ipv6-multicast\\prefix-list")
TestFlag.okfail("peer\\ipv4-unicast\\prefix-list")
TestFlag.okfail("peer\\ipv4-multicast\\prefix-list")
TestFlag.okfail("peer\\ipv6-unicast\\prefix-list")
TestFlag.okfail("peer\\ipv6-multicast\\prefix-list")
TestFlag.okfail("peer\\ipv4-unicast\\remove-private-AS")
TestFlag.okfail("peer\\ipv4-multicast\\remove-private-AS")
TestFlag.okfail("peer\\ipv6-unicast\\remove-private-AS")
TestFlag.okfail("peer\\ipv6-multicast\\remove-private-AS")
TestFlag.okfail("peer\\ipv4-unicast\\remove-private-AS all")
TestFlag.okfail("peer\\ipv4-multicast\\remove-private-AS all")
TestFlag.okfail("peer\\ipv6-unicast\\remove-private-AS all")
TestFlag.okfail("peer\\ipv6-multicast\\remove-private-AS all")
TestFlag.okfail("peer\\ipv4-unicast\\remove-private-AS replace-AS")
TestFlag.okfail("peer\\ipv4-multicast\\remove-private-AS replace-AS")
TestFlag.okfail("peer\\ipv6-unicast\\remove-private-AS replace-AS")
TestFlag.okfail("peer\\ipv6-multicast\\remove-private-AS replace-AS")
TestFlag.okfail("peer\\ipv4-unicast\\remove-private-AS all replace-AS")
TestFlag.okfail("peer\\ipv4-multicast\\remove-private-AS all replace-AS")
TestFlag.okfail("peer\\ipv6-unicast\\remove-private-AS all replace-AS")
TestFlag.okfail("peer\\ipv6-multicast\\remove-private-AS all replace-AS")
TestFlag.okfail("peer\\ipv4-unicast\\route-map")
TestFlag.okfail("peer\\ipv4-multicast\\route-map")
TestFlag.okfail("peer\\ipv6-unicast\\route-map")
TestFlag.okfail("peer\\ipv6-multicast\\route-map")
TestFlag.okfail("peer\\ipv4-unicast\\route-map")
TestFlag.okfail("peer\\ipv4-multicast\\route-map")
TestFlag.okfail("peer\\ipv6-unicast\\route-map")
TestFlag.okfail("peer\\ipv6-multicast\\route-map")
TestFlag.okfail("peer\\ipv4-unicast\\route-reflector-client")
TestFlag.okfail("peer\\ipv4-multicast\\route-reflector-client")
TestFlag.okfail("peer\\ipv6-unicast\\route-reflector-client")
TestFlag.okfail("peer\\ipv6-multicast\\route-reflector-client")
TestFlag.okfail("peer\\ipv4-unicast\\route-server-client")
TestFlag.okfail("peer\\ipv4-multicast\\route-server-client")
TestFlag.okfail("peer\\ipv6-unicast\\route-server-client")
TestFlag.okfail("peer\\ipv6-multicast\\route-server-client")
TestFlag.okfail("peer\\ipv4-unicast\\send-community")
TestFlag.okfail("peer\\ipv4-multicast\\send-community")
TestFlag.okfail("peer\\ipv6-unicast\\send-community")
TestFlag.okfail("peer\\ipv6-multicast\\send-community")
TestFlag.okfail("peer\\ipv4-unicast\\send-community extended")
TestFlag.okfail("peer\\ipv4-multicast\\send-community extended")
TestFlag.okfail("peer\\ipv6-unicast\\send-community extended")
TestFlag.okfail("peer\\ipv6-multicast\\send-community extended")
TestFlag.okfail("peer\\ipv4-unicast\\send-community large")
TestFlag.okfail("peer\\ipv4-multicast\\send-community large")
TestFlag.okfail("peer\\ipv6-unicast\\send-community large")
TestFlag.okfail("peer\\ipv6-multicast\\send-community large")
TestFlag.okfail("peer\\ipv4-unicast\\soft-reconfiguration inbound")
TestFlag.okfail("peer\\ipv4-multicast\\soft-reconfiguration inbound")
TestFlag.okfail("peer\\ipv6-unicast\\soft-reconfiguration inbound")
TestFlag.okfail("peer\\ipv6-multicast\\soft-reconfiguration inbound")
TestFlag.okfail("peer\\ipv4-unicast\\unsuppress-map")
TestFlag.okfail("peer\\ipv4-multicast\\unsuppress-map")
TestFlag.okfail("peer\\ipv6-unicast\\unsuppress-map")
TestFlag.okfail("peer\\ipv6-multicast\\unsuppress-map")
TestFlag.okfail("peer\\ipv4-unicast\\weight")
TestFlag.okfail("peer\\ipv4-multicast\\weight")
TestFlag.okfail("peer\\ipv6-unicast\\weight")
TestFlag.okfail("peer\\ipv6-multicast\\weight")
TestFlag.okfail("peer\\ipv4-vpn\\accept-own")
TestFlag.okfail("peer\\ipv6-vpn\\accept-own")



#!/usr/bin/env python
# SPDX-License-Identifier: ISC
#
# Copyright (c) 2022 by
# Donald Sharp
#

"""
Test some bgp interface based issues that show up
"""

import os
import sys
import json
import pytest
import functools

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen
from lib.common_config import step

pytestmark = [pytest.mark.bgpd]


def build_topo(tgen):
    tgen.add_router("r1")
    tgen.add_router("r2")

    switch = tgen.add_switch("s1")
    switch.add_link(tgen.gears["r1"])
    switch.add_link(tgen.gears["r2"])


def setup_module(mod):
    tgen = Topogen(build_topo, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


#
# Test these events:
# a) create an unnumbered neighbor
# b) shutdown the interface
# c) remove the unnumbered peer in bgp and bgp does not crash
def test_bgp_unnumbered_removal():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    def _bgp_nexthop_cache():
        output = tgen.gears["r1"].vtysh_cmd("show bgp nexthop")
        expected = "Current BGP nexthop cache:\n"
        return output == expected

    def _bgp_converge():
        output = json.loads(
            tgen.gears["r1"].vtysh_cmd("show ip bgp 172.16.255.254/32 json")
        )
        expected = {"prefix": "172.16.255.254/32"}

        return topotest.json_cmp(output, expected)

    step("Ensure Convergence of BGP")
    test_func = functools.partial(_bgp_converge)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=1)

    assert result is None, 'Failed bgp convergence in "{}"'.format(tgen.gears["r2"])

    step("Shutdown interface r1-eth0")

    tgen.gears["r1"].vtysh_cmd(
        """
           configure
           int r1-eth0
             shutdown
        """
    )

    step("Remove the neighbor from r1")
    tgen.gears["r1"].vtysh_cmd(
        """
           configure
           router bgp
            no neighbor r1-eth0 interface remote-as external
       """
    )

    step("Ensure that BGP does not crash")
    test_func = functools.partial(_bgp_nexthop_cache)
    _, result = topotest.run_and_expect(test_func, True, count=10, wait=1)

    assert result is True, "BGP did not crash on r1"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



# SPDX-License-Identifier: GPL-2.0-or-later
# check that the first header included in C files is either
# zebra.h or config.h
#
# Copyright (C) 2020  David Lamparter for NetDEF, Inc.

import sys
import os
import re
import subprocess
import argparse

argp = argparse.ArgumentParser(description="include fixer")
argp.add_argument("--autofix", action="store_const", const=True)
argp.add_argument("--warn-empty", action="store_const", const=True)
argp.add_argument("--pipe", action="store_const", const=True)

include_re = re.compile(r'^#\s*include\s+["<]([^ ">]+)[">]', re.M)

ignore = [
    lambda fn: fn.startswith("tools/"),
    lambda fn: fn
    in [
        "lib/elf_py.c",
    ],
]


def run(args):
    out = []

    files = subprocess.check_output(["git", "ls-files"]).decode("ASCII")
    for fn in files.splitlines():
        if not fn.endswith(".c"):
            continue
        if max([i(fn) for i in ignore]):
            continue

        with open(fn, "r") as fd:
            data = fd.read()

        m = include_re.search(data)
        if m is None:
            if args.warn_empty:
                sys.stderr.write("no #include in %s?\n" % (fn))
            continue
        if m.group(1) in ["config.h", "zebra.h", "lib/zebra.h"]:
            continue

        if args.autofix:
            sys.stderr.write("%s: %s - fixing\n" % (fn, m.group(0)))
            if fn.startswith("pceplib/"):
                insert = '#ifdef HAVE_CONFIG_H\n#include "config.h"\n#endif\n\n'
            else:
                insert = "#include <zebra.h>\n\n"

            pos = m.span()[0]

            data = data[:pos] + insert + data[pos:]
            with open(fn + ".new", "w") as fd:
                fd.write(data)
            os.rename(fn + ".new", fn)
        else:
            sys.stderr.write("%s: %s\n" % (fn, m.group(0)))
            out.append(fn)

    if len(out):
        if args.pipe:
            # for "vim `firstheader.py`"
            print("\n".join(out))
        return 1
    return 0


if __name__ == "__main__":
    args = argp.parse_args()
    sys.exit(run(args))



# SPDX-License-Identifier: GPL-2.0-or-later
#
# Test helpers for FRR
#
# Copyright (C) 2017 by David Lamparter & Christian Franke,
#                       Open Source Routing / NetDEF Inc.
#
# This file is part of FRRouting (FRR)
#

import subprocess
import sys
import re
import inspect
import os
import difflib

import frrsix

#
# These are the gritty internals of the TestMultiOut implementation.
# See below for the definition of actual TestMultiOut tests.
#

srcbase = os.path.abspath(inspect.getsourcefile(frrsix))
for i in range(0, 3):
    srcbase = os.path.dirname(srcbase)


def binpath(srcpath):
    return os.path.relpath(os.path.abspath(srcpath), srcbase)


class MultiTestFailure(Exception):
    pass


class MetaTestMultiOut(type):
    def __getattr__(cls, name):
        if name.startswith("_"):
            raise AttributeError

        internal_name = "_{}".format(name)
        if internal_name not in dir(cls):
            raise AttributeError

        def registrar(*args, **kwargs):
            cls._add_test(getattr(cls, internal_name), *args, **kwargs)

        return registrar


@frrsix.add_metaclass(MetaTestMultiOut)
class _TestMultiOut(object):
    def _run_tests(self):
        if "tests_run" in dir(self.__class__) and self.tests_run:
            return
        self.__class__.tests_run = True
        basedir = os.path.dirname(inspect.getsourcefile(type(self)))
        program = os.path.join(basedir, self.program)
        proc = subprocess.Popen([binpath(program)], stdout=subprocess.PIPE)
        self.output, _ = proc.communicate("")
        self.exitcode = proc.wait()

        self.__class__.testresults = {}
        for test in self.tests:
            try:
                test(self)
            except MultiTestFailure:
                self.testresults[test] = sys.exc_info()
            else:
                self.testresults[test] = None

    def _exit_cleanly(self):
        if self.exitcode != 0:
            raise MultiTestFailure("Program did not terminate with exit code 0")

    @classmethod
    def _add_test(cls, method, *args, **kwargs):
        if "tests" not in dir(cls):
            setattr(cls, "tests", [])
            if method is not cls._exit_cleanly:
                cls._add_test(cls._exit_cleanly)

        def matchfunction(self):
            method(self, *args, **kwargs)

        cls.tests.append(matchfunction)

        def testfunction(self):
            self._run_tests()
            result = self.testresults[matchfunction]
            if result is not None:
                frrsix.reraise(*result)

        testname = re.sub(r"[^A-Za-z0-9]", "_", "%r%r" % (args, kwargs))
        testname = re.sub(r"__*", "_", testname)
        testname = testname.strip("_")
        if not testname:
            testname = method.__name__.strip("_")
        if "test_%s" % testname in dir(cls):
            index = 2
            while "test_%s_%d" % (testname, index) in dir(cls):
                index += 1
            testname = "%s_%d" % (testname, index)
        setattr(cls, "test_%s" % testname, testfunction)


#
# This class houses the actual TestMultiOut tests types.
# If you want to add a new test type, you probably do it here.
#
# Say you want to add a test type called foobarlicious. Then define
# a function _foobarlicious here that takes self and the test arguments
# when called. That function should check the output in self.output
# to see whether it matches the expectation of foobarlicious with the
# given arguments and should then adjust self.output according to how
# much output it consumed.
# If the output doesn't meet the expectations, MultiTestFailure can be
# raised, however that should only be done after self.output has been
# modified according to consumed content.
#

re_okfail = re.compile(r"(?:[3[12]m|^)?(?P<ret>OK|failed)".encode("utf8"), re.MULTILINE)


class TestMultiOut(_TestMultiOut):
    def _onesimple(self, line):
        if type(line) is str:
            line = line.encode("utf8")
        idx = self.output.find(line)
        if idx != -1:
            self.output = self.output[idx + len(line) :]
        else:
            raise MultiTestFailure("%r could not be found" % line)

    def _okfail(self, line, okfail=re_okfail):
        self._onesimple(line)

        m = okfail.search(self.output)
        if m is None:
            raise MultiTestFailure("OK/fail not found")
        self.output = self.output[m.end() :]

        if m.group("ret") != "OK".encode("utf8"):
            raise MultiTestFailure("Test output indicates failure")


#
# This class implements a test comparing the output of a program against
# an existing reference output
#


class TestRefMismatch(Exception):
    def __init__(self, _test, outtext, reftext):
        self.outtext = outtext
        self.reftext = reftext

    def __str__(self):
        rv = "Expected output and actual output differ:\n"
        rv += "\n".join(
            difflib.unified_diff(
                self.reftext.splitlines(),
                self.outtext.splitlines(),
                "reftext",
                "outtext",
                lineterm="",
            )
        )
        return rv


class TestExitNonzero(Exception):
    pass


class TestRefOut(object):
    def test_refout(self):
        basedir = os.path.dirname(inspect.getsourcefile(type(self)))
        program = os.path.join(basedir, self.program)

        if getattr(self, "built_refin", False):
            refin = binpath(program) + ".in"
        else:
            refin = program + ".in"
        if getattr(self, "built_refout", False):
            refout = binpath(program) + ".refout"
        else:
            refout = program + ".refout"

        intext = ""
        if os.path.exists(refin):
            with open(refin, "rb") as f:
                intext = f.read()
        with open(refout, "rb") as f:
            reftext = f.read()

        proc = subprocess.Popen(
            [binpath(program)], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        outtext, _ = proc.communicate(intext)

        # Get rid of newline problems (Windows vs Unix Style)
        outtext_str = outtext.decode("utf8").replace("\r\n", "\n").replace("\r", "\n")
        reftext_str = reftext.decode("utf8").replace("\r\n", "\n").replace("\r", "\n")

        if outtext_str != reftext_str:
            raise TestRefMismatch(self, outtext_str, reftext_str)
        if proc.wait() != 0:
            raise TestExitNonzero(self)



# SPDX-License-Identifier: GPL-2.0-or-later
# some basic tests for xrelfo & the python ELF machinery
#
# Copyright (C) 2020  David Lamparter for NetDEF, Inc.

import sys
import os
import pytest
from pprint import pprint

root = os.path.dirname(os.path.dirname(__file__))
sys.path.append(os.path.join(root, "python"))

import xrelfo
from clippy import elf, uidhash


def test_uidhash():
    assert uidhash.uidhash("lib/test_xref.c", "logging call", 3, 0) == "H7KJB-67TBH"


def test_xrelfo_other():
    for data in [
        elf.ELFNull(),
        elf.ELFUnresolved("somesym", 0),
    ]:

        dissect = xrelfo.XrefPtr(data)
        print(repr(dissect))

        with pytest.raises(AttributeError):
            dissect.xref


def test_xrelfo_obj():
    xrelfo_ = xrelfo.Xrelfo()
    edf = xrelfo_.load_elf(os.path.join(root, "lib/.libs/zclient.o"), "zclient.lo")
    xrefs = xrelfo_._xrefs

    with pytest.raises(elf.ELFAccessError):
        edf[0:4]

    pprint(xrefs[0])
    pprint(xrefs[0]._data)


def test_xrelfo_bin():
    xrelfo_ = xrelfo.Xrelfo()
    edf = xrelfo_.load_elf(os.path.join(root, "lib/.libs/libfrr.so"), "libfrr.la")
    xrefs = xrelfo_._xrefs

    assert edf[0:4] == b"\x7fELF"

    pprint(xrefs[0])
    pprint(xrefs[0]._data)



#!/usr/bin/python
# SPDX-License-Identifier: ISC

#
# Copyright (c) 2021 by VMware, Inc. ("VMware")
# Used Copyright (c) 2018 by Network Device Education Foundation, Inc.
# ("NetDEF") in this file.
#


"""OSPF Basic Functionality Automation."""
import os
import sys
import time
import pytest

# Save the Current Working Directory to find configuration files.
CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))
sys.path.append(os.path.join(CWD, "../lib/"))

# pylint: disable=C0413
# Import topogen and topotest helpers
from lib.topogen import Topogen, get_topogen

# Import topoJson from lib, to create topology and initial configuration
from lib.common_config import (
    start_topology,
    write_test_header,
    write_test_footer,
    reset_config_on_routers,
    verify_rib,
    create_static_routes,
    step,
    get_frr_ipv6_linklocal,
)
from lib.topolog import logger
from lib.topojson import build_config_from_json

from lib.ospf import (
    verify_ospf6_neighbor,
    clear_ospf,
    verify_ospf6_rib,
    create_router_ospf,
)


pytestmark = [pytest.mark.ospfd, pytest.mark.staticd]


# Global variables
topo = None

NETWORK = {
    "ipv4": [
        "11.0.20.1/32",
        "11.0.20.2/32",
        "11.0.20.3/32",
        "11.0.20.4/32",
        "11.0.20.5/32",
    ],
    "ipv6": ["2::1/128", "2::2/128", "2::3/128", "2::4/128", "2::5/128"],
}
MASK = {"ipv6": "32", "ipv6": "128"}

"""
TOPOOLOGY =
      Please view in a fixed-width font such as Courier.
      Topo : Broadcast Networks
        +---+       +---+          +---+           +---+
        |R0 +       +R1 +          +R2 +           +R3 |
        +-+-+       +-+-+          +-+-+           +-+-+
          |           |              |               |
          |           |              |               |
        --+-----------+--------------+---------------+-----
                         Ethernet Segment

TESTCASES =
1.  Verify OSPF ECMP with max path configured as 8
    (Edge having 1 uplink port as broadcast network,
    connect to 8 TORs - LAN case)

 """


def setup_module(mod):
    """
    Sets up the pytest environment

    * `mod`: module name
    """
    global topo, switch_name
    testsuite_run_time = time.asctime(time.localtime(time.time()))
    logger.info("Testsuite start time: {}".format(testsuite_run_time))
    logger.info("=" * 40)

    logger.info("Running setup_module to create topology")

    # This function initiates the topology build with Topogen...
    json_file = "{}/ospfv3_ecmp_lan.json".format(CWD)
    tgen = Topogen(json_file, mod.__name__)
    global topo
    topo = tgen.json_topo
    # ... and here it calls Mininet initialization functions.

    # Starting topology, create tmp files which are loaded to routers
    #  to start daemons and then start routers
    start_topology(tgen)

    # Creating configuration from JSON
    build_config_from_json(tgen, topo)

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)
    # Api call verify whether OSPF is converged
    ospf_covergence = verify_ospf6_neighbor(tgen, topo, lan=True)
    assert ospf_covergence is True, "setup_module :Failed \n Error:  {}".format(
        ospf_covergence
    )

    switch_name = [k for k in topo["switches"].keys()][0]

    logger.info("Running setup_module() done")


def teardown_module():
    """
    Teardown the pytest environment.

    * `mod`: module name
    """

    logger.info("Running teardown_module to delete topology")

    tgen = get_topogen()

    # Stop toplogy and Remove tmp files

    try:
        # Stop toplogy and Remove tmp files
        tgen.stop_topology()

    except OSError:
        # OSError exception is raised when mininet tries to stop switch
        # though switch is stopped once but mininet tries to stop same
        # switch again, where it ended up with exception
        pass
    logger.info(
        "Testsuite end time: {}".format(time.asctime(time.localtime(time.time())))
    )
    logger.info("=" * 40)


def red_static(dut, config=True):
    """Local def for Redstribute static routes inside ospf."""
    global topo
    tgen = get_topogen()
    if config:
        ospf_red = {dut: {"ospf6": {"redistribute": [{"redist_type": "static"}]}}}
    else:
        ospf_red = {
            dut: {
                "ospf6": {"redistribute": [{"redist_type": "static", "delete": True}]}
            }
        }
    result = create_router_ospf(tgen, topo, ospf_red)
    assert result is True, "Testcase : Failed \n Error: {}".format(result)


def red_connected(dut, config=True):
    """Local def for Redstribute connected routes inside ospf."""
    global topo
    tgen = get_topogen()
    if config:
        ospf_red = {dut: {"ospf6": {"redistribute": [{"redist_type": "connected"}]}}}
    else:
        ospf_red = {
            dut: {
                "ospf6": {
                    "redistribute": [{"redist_type": "connected", "del_action": True}]
                }
            }
        }
    result = create_router_ospf(tgen, topo, ospf_red)
    assert result is True, "Testcase: Failed \n Error: {}".format(result)


def get_llip(onrouter, intf):
    """
    API to get the link local ipv6 address of a particular interface

    Parameters
    ----------
    * `fromnode`: Source node
    * `tonode` : interface for which link local ip needs to be returned.

    Usage
    -----
    result = get_llip('r1', 'r2-link0')

    Returns
    -------
    1) link local ipv6 address from the interface.
    2) errormsg - when link local ip not found.
    """
    tgen = get_topogen()
    intf = topo["routers"][onrouter]["links"][intf]["interface"]
    llip = get_frr_ipv6_linklocal(tgen, onrouter, intf)
    if llip:
        logger.info("llip ipv6 address to be set as NH is %s", llip)
        return llip
    return None


def get_glipv6(onrouter, intf):
    """
    API to get the global ipv6 address of a particular interface

    Parameters
    ----------
    * `onrouter`: Source node
    * `intf` : interface for which link local ip needs to be returned.

    Usage
    -----
    result = get_glipv6('r1', 'r2-link0')

    Returns
    -------
    1) global ipv6 address from the interface.
    2) errormsg - when link local ip not found.
    """
    glipv6 = (topo["routers"][onrouter]["links"][intf]["ipv6"]).split("/")[0]
    if glipv6:
        logger.info("Global ipv6 address to be set as NH is %s", glipv6)
        return glipv6
    return None


# ##################################
# Test cases start here.
# ##################################


def test_ospfv3_lan_ecmp_tc18_p0(request):
    """
    OSPF ECMP.

    Verify OSPF ECMP with max path configured as 8
    (Edge having 1 uplink port as broadcast network,
    connect to 8 TORs - LAN case)

    """
    tc_name = request.node.name
    write_test_header(tc_name)
    tgen = get_topogen()

    # Don't run this test if we have any failure.
    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    global topo
    step("Bring up the base config as per the topology")
    step(". Configure ospf in all the routers on LAN interface.")

    reset_config_on_routers(tgen)

    step("Verify that OSPF is up with 8 neighborship sessions.")

    ospf_covergence = verify_ospf6_neighbor(tgen, topo, lan=True)
    assert ospf_covergence is True, "Testcase Failed \n Error:  {}".format(
        ospf_covergence
    )

    step(
        "Configure a static route in all the routes and "
        "redistribute static/connected in OSPF."
    )

    for rtr in topo["routers"]:
        input_dict = {
            rtr: {
                "static_routes": [
                    {"network": NETWORK["ipv6"][0], "no_of_ip": 5, "next_hop": "Null0"}
                ]
            }
        }
        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

        dut = rtr
        red_static(dut)

    step(
        "Verify that route in R0 in stalled with 8 hops. "
        "Verify ospf route table and ip route table."
    )

    nh = []
    for rtr in topo["routers"]:
        llip = get_llip(rtr, switch_name)
        assert llip is not None, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )
        nh.append(llip)

    llip = get_llip("r1", switch_name)
    assert llip is not None, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    nh.remove(llip)
    dut = "r1"
    result = verify_ospf6_rib(tgen, dut, input_dict)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    protocol = "ospf6"
    result = verify_rib(tgen, "ipv6", dut, input_dict, protocol=protocol)
    assert result is True, "Testcase {} : Failed \n Error: {}".format(tc_name, result)

    step(" clear ip ospf interface on DUT(r0)")
    clear_ospf(tgen, "r0", ospf="ospf6")

    step(
        "Verify that after clearing the ospf interface all the "
        "neighbours are up and routes are installed with 8 next hop "
        "in ospf and ip route tables on R0"
    )

    dut = "r0"
    ospf_covergence = verify_ospf6_neighbor(tgen, topo, dut=dut, lan=True)
    assert ospf_covergence is True, "Testcase Failed \n Error:  {}".format(
        ospf_covergence
    )

    step(" clear ip ospf interface on R2")
    clear_ospf(tgen, "r2")

    dut = "r2"
    ospf_covergence = verify_ospf6_neighbor(tgen, topo, dut=dut, lan=True)
    assert ospf_covergence is True, "Testcase Failed \n Error:  {}".format(
        ospf_covergence
    )

    step("Delete static/connected cmd in ospf in all the routes one by one.")
    for rtr in topo["routers"]:
        input_dict = {
            rtr: {
                "static_routes": [
                    {
                        "network": NETWORK["ipv6"][0],
                        "no_of_ip": 5,
                        "next_hop": "Null0",
                        "delete": True,
                    }
                ]
            }
        }
        result = create_static_routes(tgen, input_dict)
        assert result is True, "Testcase {} : Failed \n Error: {}".format(
            tc_name, result
        )

    write_test_footer(tc_name)


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))



#!/usr/bin/env python
# SPDX-License-Identifier: ISC

# Copyright (c) 2022 by
# Donatas Abraitis <donatas@opensourcerouting.org>
#

"""
Test if BGP confederation works properly when using
remote-as internal/external.

Also, check if the same works with peer-groups as well.
"""

import os
import sys
import json
import pytest
import functools

pytestmark = [pytest.mark.bgpd]

CWD = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(CWD, "../"))

# pylint: disable=C0413
from lib import topotest
from lib.topogen import Topogen, TopoRouter, get_topogen


def setup_module(mod):
    topodef = {"s1": ("r1", "r2"), "s2": ("r1", "r3")}
    tgen = Topogen(topodef, mod.__name__)
    tgen.start_topology()

    router_list = tgen.routers()

    for _, (rname, router) in enumerate(router_list.items(), 1):
        router.load_config(
            TopoRouter.RD_ZEBRA, os.path.join(CWD, "{}/zebra.conf".format(rname))
        )
        router.load_config(
            TopoRouter.RD_BGP, os.path.join(CWD, "{}/bgpd.conf".format(rname))
        )

    tgen.start_router()


def teardown_module(mod):
    tgen = get_topogen()
    tgen.stop_topology()


def test_bgp_confederation_astype():
    tgen = get_topogen()

    if tgen.routers_have_failure():
        pytest.skip(tgen.errors)

    r1 = tgen.gears["r1"]

    def _bgp_converge():
        output = json.loads(r1.vtysh_cmd("show bgp summary json"))
        expected = {
            "ipv4Unicast": {
                "peerCount": 2,
                "peers": {
                    "192.168.1.2": {
                        "hostname": "r2",
                        "remoteAs": 65002,
                        "localAs": 65001,
                        "pfxRcd": 1,
                        "state": "Established",
                    },
                    "192.168.2.2": {
                        "hostname": "r3",
                        "remoteAs": 65003,
                        "localAs": 65001,
                        "pfxRcd": 1,
                        "state": "Established",
                    },
                },
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_converge)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Can't converge"

    def _bgp_check_neighbors():
        output = json.loads(r1.vtysh_cmd("show bgp neighbors json"))
        expected = {
            "192.168.1.2": {
                "nbrCommonAdmin": True,
                "nbrConfedExternalLink": True,
                "hostname": "r2",
            },
            "192.168.2.2": {
                "nbrCommonAdmin": True,
                "nbrConfedExternalLink": True,
                "hostname": "r3",
            },
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_check_neighbors)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Can't see neighbors to be in BGP confederation"

    def _bgp_check_routes():
        output = json.loads(r1.vtysh_cmd("show bgp ipv4 unicast json"))
        expected = {
            "routes": {
                "172.16.255.254/32": [
                    {
                        "valid": True,
                        "pathFrom": "external",
                        "path": "(65003)",
                    },
                    {
                        "valid": True,
                        "pathFrom": "external",
                        "path": "(65002)",
                    },
                ]
            }
        }
        return topotest.json_cmp(output, expected)

    test_func = functools.partial(_bgp_check_routes)
    _, result = topotest.run_and_expect(test_func, None, count=60, wait=0.5)
    assert result is None, "Can't see routes to be in BGP confederation"


if __name__ == "__main__":
    args = ["-s"] + sys.argv[1:]
    sys.exit(pytest.main(args))
